"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.neptune.create_db_cluster
hub.exec.boto3.client.neptune.delete_db_cluster
hub.exec.boto3.client.neptune.describe_db_clusters
hub.exec.boto3.client.neptune.failover_db_cluster
hub.exec.boto3.client.neptune.modify_db_cluster
hub.exec.boto3.client.neptune.start_db_cluster
hub.exec.boto3.client.neptune.stop_db_cluster
"""
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]
create_waiter_acceptors = [
    {
        "matcher": "pathAll",
        "expected": "available",
        "state": "success",
        "argument": "DBClusters[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "creating",
        "state": "retry",
        "argument": "DBClusters[].Status",
    },
]
update_waiter_acceptors = [
    {
        "matcher": "pathAll",
        "expected": "available",
        "state": "success",
        "argument": "DBClusters[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "modifying",
        "state": "retry",
        "argument": "DBClusters[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "resetting-master-credentials",
        "state": "retry",
        "argument": "DBClusters[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "renaming",
        "state": "retry",
        "argument": "DBClusters[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "upgrading",
        "state": "retry",
        "argument": "DBClusters[].Status",
    },
]
delete_waiter_acceptors = [
    {
        "matcher": "error",
        "expected": "DBClusterNotFoundFault",
        "state": "success",
        "argument": "Error.Code",
    },
    {
        "matcher": "pathAll",
        "expected": "available",
        "state": "retry",
        "argument": "DBClusters[].Status",
    },
    {
        "matcher": "pathAll",
        "expected": "deleting",
        "state": "retry",
        "argument": "DBClusters[].Status",
    },
]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    *,
    engine: str,
    availability_zones: List = None,
    backup_retention_period: int = None,
    copy_tags_to_snapshot: bool = None,
    database_name: str = None,
    db_cluster_parameter_group_name: str = None,
    vpc_security_group_ids: List = None,
    db_subnet_group_name: str = None,
    engine_version: str = None,
    port: int = None,
    master_username: str = None,
    master_user_password: str = None,
    option_group_name: str = None,
    preferred_backup_window: str = None,
    preferred_maintenance_window: str = None,
    replication_source_identifier: str = None,
    tags: List = None,
    storage_encrypted: bool = None,
    kms_key_id: str = None,
    enable_iam_database_authentication: bool = None,
    enable_cloudwatch_logs_exports: List = None,
    deletion_protection: bool = True,
    source_region: str = None,
    apply_immediately: bool = None,
    cloudwatch_logs_export_configuration: Dict = None,
    db_instance_parameter_group_name: str = None,
    allow_major_version_upgrades: bool = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a new Amazon Neptune DB cluster. You can use the ReplicationSourceIdentifier parameter to create the DB
    cluster as a Read Replica of another DB cluster or Amazon Neptune DB instance. Note that when you create a new
    cluster using CreateDBCluster directly, deletion protection is disabled by default (when you create a new
    production cluster in the console, deletion protection is enabled by default). You can only delete a DB cluster
    if its DeletionProtection field is set to false.

    Args:
        name(str): An Idem name of the resource. This also maps to db_cluster_identifier while creating the cluster.
        resource_id(str, optional): An identifier of the resource in the provider, this also maps to db_cluster_identifier during update. Defaults to None.
        availability_zones(List, optional): A list of EC2 Availability Zones that instances in the DB cluster can be created in. Defaults to None.
        backup_retention_period(int, optional): The number of days for which automated backups are retained. You must specify a minimum value of
            1. Default: 1 Constraints:   Must be a value from 1 to 35. Defaults to None.
        copy_tags_to_snapshot(bool, optional):  If set to true, tags are copied to any snapshot of the DB cluster that is created. Defaults to None.
        database_name(str, optional): The name for your database of up to 64 alpha-numeric characters. If you do not provide a name,
            Amazon Neptune will not create a database in the DB cluster you are creating. Defaults to None.
        db_cluster_identifier(str): The DB cluster identifier. This parameter is stored as a lowercase string. Constraints:   Must
            contain from 1 to 63 letters, numbers, or hyphens.   First character must be a letter.   Cannot
            end with a hyphen or contain two consecutive hyphens.   Example: my-cluster1.
        db_cluster_parameter_group_name(str, optional):  The name of the DB cluster parameter group to associate with this DB cluster. If this argument
            is omitted, the default is used. Constraints:   If supplied, must match the name of an existing
            DBClusterParameterGroup. Defaults to None.
        vpc_security_group_ids(List, optional): A list of EC2 VPC security groups to associate with this DB cluster. Defaults to None.
        db_subnet_group_name(str, optional): A DB subnet group to associate with this DB cluster. Constraints: Must match the name of an
            existing DBSubnetGroup. Must not be default. Example: mySubnetgroup. Defaults to None.
        engine(str): The name of the database engine to be used for this DB cluster. Valid Values: neptune.
        engine_version(str, optional): The version number of the database engine to use for the new DB cluster. Example: 1.0.2.1. Defaults to None.
        port(int, optional): The port number on which the instances in the DB cluster accept connections.  Default: 8182. Defaults to None.
        preferred_backup_window(str, optional): The daily time range during which automated backups are created if automated backups are enabled
            using the BackupRetentionPeriod parameter. The default is a 30-minute window selected at random
            from an 8-hour block of time for each Amazon Region. To see the time blocks available, see
            Adjusting the Preferred Maintenance Window in the Amazon Neptune User Guide.  Constraints:
            Must be in the format hh24:mi-hh24:mi.   Must be in Universal Coordinated Time (UTC).   Must not
            conflict with the preferred maintenance window.   Must be at least 30 minutes. Defaults to None.
        preferred_maintenance_window(str, optional): The weekly time range during which system maintenance can occur, in Universal Coordinated Time
            (UTC). Format: ddd:hh24:mi-ddd:hh24:mi  The default is a 30-minute window selected at random
            from an 8-hour block of time for each Amazon Region, occurring on a random day of the week. To
            see the time blocks available, see  Adjusting the Preferred Maintenance Window in the Amazon
            Neptune User Guide.  Valid Days: Mon, Tue, Wed, Thu, Fri, Sat, Sun. Constraints: Minimum
            30-minute window. Defaults to None.
        replication_source_identifier(str, optional): The Amazon Resource Name (ARN) of the source DB instance or DB cluster if this DB cluster is
            created as a Read Replica. Defaults to None.
        tags(List, optional): The tags to assign to the new DB cluster. Defaults to None.
        storage_encrypted(bool, optional): Specifies whether the DB cluster is encrypted. Defaults to None.
        kms_key_id(str, optional): The Amazon KMS key identifier for an encrypted DB cluster. The KMS key identifier is the Amazon
            Resource Name (ARN) for the KMS encryption key. If you are creating a DB cluster with the same
            Amazon account that owns the KMS encryption key used to encrypt the new DB cluster, then you can
            use the KMS key alias instead of the ARN for the KMS encryption key. If an encryption key is not
            specified in KmsKeyId:   If ReplicationSourceIdentifier identifies an encrypted source, then
            Amazon Neptune will use the encryption key used to encrypt the source. Otherwise, Amazon Neptune
            will use your default encryption key.   If the StorageEncrypted parameter is true and
            ReplicationSourceIdentifier is not specified, then Amazon Neptune will use your default
            encryption key.   Amazon KMS creates the default encryption key for your Amazon account. Your
            Amazon account has a different default encryption key for each Amazon Region. If you create a
            Read Replica of an encrypted DB cluster in another Amazon Region, you must set KmsKeyId to a KMS
            key ID that is valid in the destination Amazon Region. This key is used to encrypt the Read
            Replica in that Amazon Region. Defaults to None.
        enable_iam_database_authentication(bool, optional): If set to true, enables Amazon Identity and Access Management (IAM) authentication for the
            entire DB cluster (this cannot be set at an instance level). Default: false. Defaults to None.
        enable_cloudwatch_logs_exports(List, optional): The list of log types that need to be enabled for exporting to CloudWatch Logs. Defaults to None.
        deletion_protection(bool, optional): A value that indicates whether the DB cluster has deletion protection enabled. The database
            can't be deleted when deletion protection is enabled. By default, deletion protection is
            enabled. Defaults to None.
        source_region(str, optional): The ID of the region that contains the source for the db cluster. Defaults to None.
        apply_immediately(bool, optional): A value that indicates whether the modifications in this request and any
                                           pending modifications are asynchronously applied as soon as possible,
                                           regardless of the PreferredMaintenanceWindow setting for the DB cluster.
                                           If this parameter is disabled, changes to the DB cluster are applied during
                                           the next maintenance window.
                                           The ApplyImmediately parameter only affects the
                                           EnableIAMDatabaseAuthentication,MasterUserPassword and NewDBClusterIdentifier
                                           values.If the ApplyImmediately parameter is disabled, then changes to the
                                           EnableIAMDatabaseAuthentication,MasterUserPassword and NewDBClusterIdentifier
                                           values are applied during the next maintenance window. All other changes are
                                           applied immediately,regardless of the value of the ApplyImmediately parameter
        cloudwatch_logs_export_configuration(Dict, optional): The configuration setting for the log types to be enabled
                                                              for export to CloudWatch Logs for a specific DB cluster.
        db_instance_parameter_group_name(Text, optional): The name of the DB parameter group to apply to all instances of
                                                         the DB cluster.
        allow_major_version_upgrades(bool, optional): A value that indicates whether major version upgrades are allowed.
        timeout(Dict, optional): Timeout configuration for create/update of AWS DB Cluster.
            * create (Dict) -- Timeout configuration for creating DB Cluster
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- Customized timeout configuration containing delay and max attempts.
            * update (Dict) -- Timeout configuration for updating DB Cluster
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- Customized timeout configuration containing delay and max attempts.
    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws.neptune.db_cluster.present:
                - name: value
                - engine: value
                - master_username: value
                - master_user_password: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    desired_state = {
        "name": name,
        "availability_zones": availability_zones,
        "backup_retention_period": backup_retention_period,
        "copy_tags_to_snapshot": copy_tags_to_snapshot,
        "database_name": database_name,
        "db_cluster_parameter_group_name": db_cluster_parameter_group_name,
        "vpc_security_group_ids": vpc_security_group_ids,
        "db_subnet_group_name": db_subnet_group_name,
        "engine": engine,
        "engine_version": engine_version,
        "port": port,
        "master_username": master_username,
        "master_user_password": master_user_password,
        "option_group_name": option_group_name,
        "preferred_backup_window": preferred_backup_window,
        "preferred_maintenance_window": preferred_maintenance_window,
        "replication_source_identifier": replication_source_identifier,
        "tags": tags,
        "storage_encrypted": storage_encrypted,
        "kms_key_id": kms_key_id,
        "enable_iam_database_authentication": enable_iam_database_authentication,
        "enable_cloudwatch_logs_exports": enable_cloudwatch_logs_exports,
        "deletion_protection": deletion_protection,
        "source_region": source_region,
    }
    hub.log.debug(f"desired state is {desired_state}")
    is_created = False

    if not resource_id:
        # if there is no resource_id, create db cluster
        # if test run, generate the new state and set in result
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={}, desired_state=desired_state
            )
            resource_id = desired_state["name"]
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.neptune.db_cluster", name=name
            )
        else:
            # if this is not a test run, create the real cluster
            create_db_cluster_result = (
                await hub.exec.boto3.client.neptune.create_db_cluster(
                    ctx,
                    DBClusterIdentifier=name,
                    AvailabilityZones=availability_zones,
                    BackupRetentionPeriod=backup_retention_period,
                    CopyTagsToSnapshot=copy_tags_to_snapshot,
                    DatabaseName=database_name,
                    DBClusterParameterGroupName=db_cluster_parameter_group_name,
                    VpcSecurityGroupIds=vpc_security_group_ids,
                    DBSubnetGroupName=db_subnet_group_name,
                    Engine=engine,
                    EngineVersion=engine_version,
                    Port=port,
                    MasterUsername=master_username,
                    MasterUserPassword=master_user_password,
                    OptionGroupName=option_group_name,
                    PreferredBackupWindow=preferred_backup_window,
                    PreferredMaintenanceWindow=preferred_maintenance_window,
                    ReplicationSourceIdentifier=replication_source_identifier,
                    Tags=tags,
                    StorageEncrypted=storage_encrypted,
                    KmsKeyId=kms_key_id,
                    EnableIAMDatabaseAuthentication=enable_iam_database_authentication,
                    EnableCloudwatchLogsExports=enable_cloudwatch_logs_exports,
                    DeletionProtection=deletion_protection,
                    SourceRegion=source_region,
                )
            )
            # if create failed, fail the request
            if not create_db_cluster_result["result"]:
                result["comment"] += create_db_cluster_result["comment"]
                result["result"] = False
                return result

            # using the resource_id, wait till db cluster is created
            resource_id = create_db_cluster_result["ret"]["DBCluster"][
                "DBClusterIdentifier"
            ]
            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=15,
                default_max_attempts=40,
                timeout_config=timeout.get("create") if timeout else None,
            )
            cluster_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
                name="ClusterCreated",
                operation="DescribeDBClusters",
                argument=["DBClusters[].Status"],
                acceptors=create_waiter_acceptors,
                client=hub.tool.boto3.client.get_client(ctx, "neptune"),
            )
            try:
                await hub.tool.boto3.client.wait(
                    ctx,
                    "neptune",
                    "ClusterCreated",
                    cluster_waiter,
                    DBClusterIdentifier=resource_id,
                    WaiterConfig=waiter_config,
                )
                result["comment"] = hub.tool.aws.comment_utils.create_comment(
                    resource_type="aws.neptune.db_cluster", name=name
                )
            except Exception as e:
                result["comment"] += (str(e),)
                result["result"] = False
                return result
        is_created = True

    # neptune db_cluster has some parameters that can only be applied while modification,
    # so if user specifies those parameters as during creation of a new db cluster, we have to modify
    # the new db cluster to achieve desired state
    desired_state["apply_immediately"] = apply_immediately
    desired_state[
        "cloudwatch_logs_export_configuration"
    ] = cloudwatch_logs_export_configuration
    desired_state["db_instance_parameter_group_name"] = db_instance_parameter_group_name
    desired_state["allow_major_version_upgrades"] = allow_major_version_upgrades

    update_db_cluster_result = await hub.tool.aws.neptune.db_cluster.update_db_cluster(
        ctx=ctx,
        name=name,
        resource_id=resource_id,
        desired_state=desired_state,
        is_created=is_created,
        update_waiter_acceptors=update_waiter_acceptors,
        timeout=timeout,
    )
    if not (
        update_db_cluster_result
        and update_db_cluster_result["result"]
        and update_db_cluster_result["new_state"]
    ):
        result["result"] = False
        result["comment"] += update_db_cluster_result["comment"]
        return result

    result["old_state"] = update_db_cluster_result["old_state"]
    result["new_state"] = update_db_cluster_result["new_state"]
    # update the result comment if the resource is not being created
    if not is_created:
        result["comment"] += update_db_cluster_result["comment"]

    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    skip_final_snapshot: bool = None,
    final_db_snapshot_identifier: str = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    The DeleteDBCluster action deletes a previously provisioned DB cluster. When you delete a DB cluster, all
    automated backups for that DB cluster are deleted and can't be recovered. Manual DB cluster snapshots of the
    specified DB cluster are not deleted. Note that the DB Cluster cannot be deleted if deletion protection is
    enabled. To delete it, you must first set its DeletionProtection field to False.

    Args:
        name(str): An Idem name of the resource.
        resource_id(str, optional): An identifier of the resource in the provider. The DB cluster identifier for the DB cluster to be deleted. This parameter isn't case-sensitive.
            Idem automatically considers this resource being absent if this field is not specified.
            Constraints:   Must match an existing DBClusterIdentifier.
        skip_final_snapshot(bool, optional):  Determines whether a final DB cluster snapshot is created before the DB cluster is deleted. If
            true is specified, no DB cluster snapshot is created. If false is specified, a DB cluster
            snapshot is created before the DB cluster is deleted.  You must specify a
            FinalDBSnapshotIdentifier parameter if SkipFinalSnapshot is false.  Default: false. Defaults to None.
        final_db_snapshot_identifier(str, optional):  The DB cluster snapshot identifier of the new DB cluster snapshot created when
            SkipFinalSnapshot is set to false.   Specifying this parameter and also setting the
            SkipFinalShapshot parameter to true results in an error.  Constraints:   Must be 1 to 255
            letters, numbers, or hyphens.   First character must be a letter   Cannot end with a hyphen or
            contain two consecutive hyphens. Defaults to None.
        timeout(Dict, optional): Timeout configuration for create/update of AWS DB Cluster.
            * create (Dict) -- Timeout configuration for creating DB Cluster
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- Customized timeout configuration containing delay and max attempts.
            * update (Dict) -- Timeout configuration for updating DB Cluster
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- Customized timeout configuration containing delay and max attempts.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws_auto.neptune.db_cluster.absent:
                - resource_id: value
    """
    already_absent_msg = hub.tool.aws.comment_utils.already_absent_comment(
        resource_type="aws.neptune.db_cluster", name=name
    )

    result = dict(
        comment=already_absent_msg,
        old_state=None,
        new_state=None,
        name=name,
        result=True,
    )

    if not resource_id:
        return result

    before = await hub.exec.boto3.client.neptune.describe_db_clusters(
        ctx, DBClusterIdentifier=resource_id
    )
    db_clusters = before["ret"].get("DBClusters") if before.get("ret") else None

    if not (before["result"] and db_clusters):
        if not ("DBClusterNotFoundFault" in str(before["comment"])):
            result["comment"] = before["comment"]
            result["result"] = False
        return result

    resource_arn = db_clusters[0].get("DBClusterArn")
    tags = await hub.exec.aws.neptune.tag.get_tags_for_resource(
        ctx, resource_arn=resource_arn
    )
    if not tags["result"]:
        result["result"] = False
        result["comment"] = tags["comment"]
        return result
    tags = tags["ret"]
    result[
        "old_state"
    ] = await hub.tool.aws.neptune.db_cluster.convert_raw_db_cluster_to_present(
        db_clusters[0], idem_resource_name=resource_id, tags=tags
    )

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.neptune.db_cluster", name=name
        )
        return result

    ret = await hub.exec.boto3.client.neptune.delete_db_cluster(
        ctx,
        DBClusterIdentifier=resource_id,
        SkipFinalSnapshot=skip_final_snapshot,
        FinalDBSnapshotIdentifier=final_db_snapshot_identifier,
    )
    result["result"] = ret["result"]
    if not result["result"]:
        result["comment"] = ret["comment"]
        return result
    # Custom waiter for delete
    waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
        default_delay=15,
        default_max_attempts=40,
        timeout_config=timeout.get("delete") if timeout else None,
    )
    cluster_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
        name="ClusterDelete",
        operation="DescribeDBClusters",
        argument=["Error.Code", "DBClusters[].Status"],
        acceptors=delete_waiter_acceptors,
        client=hub.tool.boto3.client.get_client(ctx, "neptune"),
    )
    try:
        await hub.tool.boto3.client.wait(
            ctx,
            "neptune",
            "ClusterDelete",
            cluster_waiter,
            DBClusterIdentifier=resource_id,
            WaiterConfig=waiter_config,
        )
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.neptune.db_cluster", name=name
        )
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns information about provisioned Neptune DB clusters.  This operation can also return
    information for Amazon RDS clusters and Amazon DocDB clusters.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.neptune.db_cluster
    """

    result = {}
    ret = await hub.exec.boto3.client.neptune.describe_db_clusters(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe aws.neptune.db_cluster {ret['comment']}")
        return {}

    for db_cluster in ret["ret"]["DBClusters"]:
        resource_id = db_cluster.get("DBClusterIdentifier")
        resource_arn = db_cluster.get("DBClusterArn")
        tags = await hub.exec.aws.neptune.tag.get_tags_for_resource(
            ctx, resource_arn=resource_arn
        )
        if not tags["result"]:
            # if something goes wrong fetching the tags (not if the tags are None that is a normal path)
            hub.log.warning(
                f"Failed listing tags for aws.neptune.db_cluster '{resource_id}'"
                f"Describe will skip this aws.neptune.db_cluster and continue. "
            )
            continue
        tags = tags["ret"]
        # Including fields to match the 'present' function parameters
        resource_translated = (
            await hub.tool.aws.neptune.db_cluster.convert_raw_db_cluster_to_present(
                db_cluster, idem_resource_name=resource_id, tags=tags
            )
        )
        result[resource_id] = {
            "aws.neptune.db_cluster.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



from collections import OrderedDict
from typing import Any
from typing import Dict
from typing import List

"""
Util functions to convert raw Dynamo DB resource state from AWS to present input format.
"""


def convert_raw_dynamodb_table_to_present(
    hub,
    raw_resource: Dict[str, Any],
    idem_resource_name: str = None,
    tags: List = None,
) -> [str, Any]:
    describe_parameters = OrderedDict(
        {
            "AttributeDefinitions": "attribute_definitions",
            "KeySchema": "key_schema",
            "TableStatus": "table_status",
        }
    )
    new_table_resource = {"resource_id": idem_resource_name, "name": idem_resource_name}
    for parameter_old_key, parameter_new_key in describe_parameters.items():
        if raw_resource.get(parameter_old_key) is not None:
            new_table_resource[parameter_new_key] = raw_resource.get(parameter_old_key)
    if raw_resource.get("StreamSpecification"):
        new_table_resource["stream_specification"] = raw_resource.get(
            "StreamSpecification"
        ).copy()
    if raw_resource.get("BillingModeSummary"):
        new_table_resource["billing_mode"] = raw_resource.get("BillingModeSummary").get(
            "BillingMode"
        )
    if raw_resource.get("TableClassSummary"):
        new_table_resource["table_class"] = raw_resource.get("TableClassSummary").get(
            "TableClass"
        )
    if raw_resource.get("LocalSecondaryIndexes"):
        local_secondary_indexes = []
        for local_index in raw_resource.get("LocalSecondaryIndexes"):
            new_local_index = {
                "IndexName": local_index["IndexName"],
                "KeySchema": local_index["KeySchema"],
                "Projection": local_index["Projection"],
            }
            local_secondary_indexes.append(new_local_index)
        new_table_resource["local_secondary_indexes"] = local_secondary_indexes

    if raw_resource.get("GlobalSecondaryIndexes"):
        global_secondary_indexes = []
        for global_index in raw_resource.get("GlobalSecondaryIndexes"):
            new_global_index = {
                "IndexName": global_index["IndexName"],
                "KeySchema": global_index["KeySchema"],
                "Projection": global_index["Projection"],
            }
            if global_index.get("ProvisionedThroughput"):
                provisioned_throughput = global_index.get("ProvisionedThroughput")
                new_provisioned = {
                    "ReadCapacityUnits": provisioned_throughput["ReadCapacityUnits"],
                    "WriteCapacityUnits": provisioned_throughput["WriteCapacityUnits"],
                }
                new_global_index["ProvisionedThroughput"] = new_provisioned
            global_secondary_indexes.append(new_global_index)
        new_table_resource["global_secondary_indexes"] = global_secondary_indexes

    if raw_resource.get("ProvisionedThroughput"):
        provisioned_throughput = raw_resource.get("ProvisionedThroughput")
        new_provisioned = {
            "ReadCapacityUnits": provisioned_throughput["ReadCapacityUnits"],
            "WriteCapacityUnits": provisioned_throughput["WriteCapacityUnits"],
        }
        new_table_resource["provisioned_throughput"] = new_provisioned

    if raw_resource.get("SSEDescription"):
        sse_specification = raw_resource.get("SSEDescription")
        new_sse_specification = {
            "KMSMasterKeyArn": sse_specification["KMSMasterKeyArn"],
            "SSEType": sse_specification["SSEType"],
        }
        if (
            sse_specification.get("Status") == "ENABLED"
            or sse_specification.get("Status") == "ENABLING"
        ):
            new_sse_specification["Enabled"] = True
        else:
            new_sse_specification["Enabled"] = False
        new_table_resource["sse_specification"] = new_sse_specification
    if raw_resource.get("Replicas"):
        new_table_resource["replica_updates"] = raw_resource.get("Replicas")
    if tags:
        new_table_resource["tags"] = hub.tool.aws.tag_utils.convert_tag_list_to_dict(
            tags
        )
    return new_table_resource



import copy
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple


def convert_tag_list_to_dict(hub, tags: List):
    result = {}
    for tag in tags:
        if "Key" in tag:
            result[tag["Key"]] = tag.get("Value")
        else:
            hub.log.warning(
                f"tag {tag} is not in the proper format of 'Key', 'Value' pair"
            )
    return result


def convert_tag_dict_to_list(hub, tags: Dict) -> List:
    if tags is None:
        tags = {}
    return [{"Key": key, "Value": value} for key, value in tags.items()]


def diff_tags_list(hub, old_tags: List, new_tags: List) -> Tuple:
    """

    Args:
        hub:
        old_tags:
        new_tags:

    Returns:

    """
    if old_tags is None:
        old_tags = []
    if new_tags is None:
        new_tags = []
    old_tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(old_tags)
    new_tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(new_tags)
    tags_to_remove, tags_to_add = hub.tool.aws.tag_utils.diff_tags_dict(
        old_tags, new_tags
    )
    return [{"Key": key, "Value": value} for key, value in tags_to_remove.items()], [
        {"Key": key, "Value": value} for key, value in tags_to_add.items()
    ]


def diff_tags_dict(hub, old_tags: Dict[str, Any], new_tags: Dict[str, Any]) -> Tuple:
    """

    Args:
        hub:
        old_tags:
        new_tags:

    Returns:

    """
    tags_to_add = {}
    tags_to_remove = {}
    new_tags = copy.deepcopy(new_tags)
    if old_tags is None:
        old_tags = {}
    if new_tags is None:
        new_tags = {}
    for key, value in old_tags.items():
        if key in new_tags:
            if old_tags[key] != new_tags[key]:
                tags_to_remove.update({key: old_tags[key]})
                tags_to_add.update({key: new_tags[key]})
            new_tags.pop(key)
        else:
            tags_to_remove.update({key: old_tags[key]})
    tags_to_add.update(new_tags)
    return tags_to_remove, tags_to_add


def convert_auto_scaling_tag_list_to_dict(hub, tags: List):
    result = {}
    for tag in tags:
        if "key" in tag:
            result[tag["key"]] = tag["value"]
        if "propagate_at_launch" in tag:
            result[f"propagate_at_launch-{tag['key']}"] = tag["propagate_at_launch"]
    return result


def convert_auto_scaling_tag_dict_to_list(hub, tags: Dict):
    result = []
    if tags is None:
        tags = {}
    for key, value in tags.items():
        if "propagate_at_launch" == key.split("-", 1)[0] and isinstance(value, bool):
            continue
        temp_tag = {"key": key, "value": value}
        if f"propagate_at_launch-{key}" in tags:
            temp_tag["propagate_at_launch"] = tags.get(f"propagate_at_launch-{key}")
        result.append(temp_tag)
    return result


def diff_auto_scaling_dict_tags(
    hub, old_tags: Dict[str, Any], new_tags: Dict[str, Any]
):
    tags_to_add = {}
    tags_to_remove = {}
    new_tags = copy.deepcopy(new_tags)
    if old_tags is None:
        old_tags = {}
    if new_tags is None:
        new_tags = {}
    for key, value in old_tags.items():
        if "propagate_at_launch" == key.split("-", 1)[0] and isinstance(value, bool):
            continue
        if key in new_tags:
            if old_tags[key] != new_tags[key]:
                tags_to_remove.update({key: old_tags[key]})
                tags_to_add.update({key: new_tags[key]})
                if f"propagate_at_launch-{key}" in old_tags:
                    tags_to_remove.update(
                        {
                            f"propagate_at_launch-{key}": old_tags[
                                f"propagate_at_launch-{key}"
                            ]
                        }
                    )
                if f"propagate_at_launch-{key}" in new_tags:
                    tags_to_add.update(
                        {
                            f"propagate_at_launch-{key}": new_tags[
                                f"propagate_at_launch-{key}"
                            ]
                        }
                    )
            elif (
                old_tags[f"propagate_at_launch-{key}"]
                != new_tags[f"propagate_at_launch-{key}"]
            ):
                tags_to_remove.update(
                    {
                        key: old_tags[key],
                        f"propagate_at_launch-{key}": old_tags[
                            f"propagate_at_launch-{key}"
                        ],
                    }
                )
                tags_to_add.update(
                    {
                        key: new_tags[key],
                        f"propagate_at_launch-{key}": new_tags[
                            f"propagate_at_launch-{key}"
                        ],
                    }
                )
            new_tags.pop(key)
            new_tags.pop(f"propagate_at_launch-{key}")
        else:
            tags_to_remove.update({key: old_tags[key]})
            if f"propagate_at_launch-{key}" in old_tags:
                tags_to_remove.update(
                    {
                        f"propagate_at_launch-{key}": old_tags[
                            f"propagate_at_launch-{key}"
                        ]
                    }
                )
    tags_to_add.update(new_tags)
    return tags_to_remove, tags_to_add



"""
Autogenerated state module using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ce.create_anomaly_monitor
hub.exec.boto3.client.ce.delete_anomaly_monitor
hub.exec.boto3.client.ce.get_anomaly_monitors
hub.exec.boto3.client.ce.update_anomaly_monitor
resource = hub.exec.boto3.client.ce.create_anomaly_monitor(ctx, AnomalyMonitor)
"""
import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    monitor_name: str,
    monitor_type: str,
    monitor_dimension: str = None,
    monitor_specification: dict = None,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a new cost anomaly detection monitor with the requested type and monitor specification.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text): Monitor ARN to identify the resource.
        monitor_name: Name of anomaly monitor.
        monitor_type(Text): Possible type values. Valid Values: DIMENSIONAL | CUSTOM
        monitor_dimension(Text): The dimensions to evaluate : SERVICE
        monitor_specification(Dict, optional): Use Expression to filter by cost or by usage.
            Defaults to None.
            * And (list) -- Return results that match both Dimension objects.
            * Or (list) -- Return results that match either Dimension object.
            * Not (list) -- Return results that don't match a Dimension object.
            * Dimensions (dict) -- The specific Dimension to use for Expression.
                ** Key (string) -- The names of the metadata types that you can use to filter and group your results. For example, AZ returns a list of Availability Zones.
                ** Values (list) -- The metadata values that you can use to filter and group your results. You can use GetDimensionValues to find specific values.
                ** MatchOptions (list) -- The match options that you can use to filter your results. MatchOptions is only applicable for actions related to Cost Category. The default values for MatchOptions are EQUALS and CASE_SENSITIVE .
            * Tags (dict) -- The specific Tag to use for Expression.
                ** Key (string) -- The key for the tag.
                ** Values (list) -- The specific value of the tag.
                ** MatchOptions (list) -- The match options that you can use to filter your results. MatchOptions is only applicable for actions related to Cost Category. The default values for MatchOptions are EQUALS and CASE_SENSITIVE .
            * CostCategories (dict) -- The filter that's based on CostCategory values.
                ** Key (string) -- The unique name of the Cost Category.
                ** Values (list) -- The specific value of the Cost Category.
                ** MatchOptions (list) -- The match options that you can use to filter your results. MatchOptions is only applicable for actions related to cost category. The default values for MatchOptions is EQUALS and CASE_SENSITIVE .

    Request Syntax:
        [monitor-resource-id]:
          aws.costexplorer.anomaly_monitor.present:
          - resource_id: 'string'
          - monitor_name: 'string'
          - monitor_type: 'string'
          - monitor_specification:
            - Dimensions: 'dict'

    Returns:
        Dict[str, str]

    Examples:

        .. code-block:: sls

            cost_monitor1234:
              aws.costexplorer.anomaly_monitor.present:
              - name: cost_monitor1234
              - resource_id: cost_monitor1234
              - monitor_specification:
                Dimensions:
                  Key: LINKED_ACCOUNT
                  Values:
                    - '820272282974'
              - monitor_type: CUSTOM
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False

    if resource_id:
        before = await hub.exec.boto3.client.ce.get_anomaly_monitors(
            ctx, MonitorArnList=[resource_id]
        )

    if before and before["ret"].get("AnomalyMonitors"):
        before = before["ret"]["AnomalyMonitors"][0]
    else:
        before = None

    if not before:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "anomaly_monitor": {
                        "monitor_name": monitor_name,
                        "monitor_specification": monitor_specification,
                        "monitor_type": monitor_type,
                        "monitor_dimension": monitor_dimension,
                    },
                },
            )
            result["comment"] = (
                f"Would create aws.costexplorer.anomaly_monitor {name}",
            )
            return result

        monitor = None
        if monitor_specification:
            monitor = {
                "MonitorName": monitor_name,
                "MonitorSpecification": monitor_specification,
                "MonitorType": monitor_type,
            }
        else:
            monitor = {
                "MonitorName": monitor_name,
                "MonitorDimension": monitor_dimension,
                "MonitorType": monitor_type,
            }
        ret = await hub.exec.boto3.client.ce.create_anomaly_monitor(
            ctx,
            AnomalyMonitor=monitor,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = result["comment"] + (
            f"Created aws.costexplorer.anomaly_monitor '{name}'",
        )
        resource_id = ret["ret"]["MonitorArn"]
    else:

        result[
            "old_state"
        ] = hub.tool.aws.costexplorer.conversion_utils.convert_raw_monitor_to_present(
            ctx, raw_resource=before, idem_resource_name=name
        )
        plan_state = copy.deepcopy(result["old_state"])
        resource_id = before.get("MonitorArn")
        update_ret = await hub.exec.aws.costexplorer.anomaly_monitor.update_monitor(
            ctx, before=before, monitor_name=monitor_name, monitor_arn=resource_id
        )
        result["comment"] = result["comment"] + update_ret["comment"]
        result["result"] = update_ret["result"]
        resource_updated = resource_updated or bool(update_ret["ret"])
        if update_ret["ret"] and ctx.get("test", False):
            if "monitor_name" in update_ret["ret"]:
                plan_state["monitor_name"] = update_ret["ret"]["monitor_name"]

        if not resource_updated:
            result["comment"] = result["comment"] + (f"{name} already exists",)

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.ce.get_anomaly_monitors(
                ctx, MonitorArnList=[resource_id]
            )
            result[
                "new_state"
            ] = hub.tool.aws.costexplorer.conversion_utils.convert_raw_monitor_to_present(
                ctx,
                raw_resource=after["ret"]["AnomalyMonitors"][0],
                idem_resource_name=name,
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(hub, ctx, name: str, resource_id: str) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes a cost anomaly monitor by the specified monitor ARN as resource_id.

    Args:
        name(Text): The Idem name of the anomaly monitor.
        resource_id(Text): Monitor ARN to identify the resource.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            cost_monitor1234:
              aws.costexplorer.anomaly_monitor.absent:
                - name: value
                - resource_id: cost_monitor1234
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.boto3.client.ce.get_anomaly_monitors(
        ctx, MonitorArnList=[resource_id]
    )

    if before and before["ret"].get("AnomalyMonitors"):
        before = before["ret"]["AnomalyMonitors"][0]
    else:
        before = None

    if not before:
        result["comment"] = (
            f"aws.costexplorer.anomaly_monitor '{name}' already absent",
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = hub.tool.aws.costexplorer.conversion_utils.convert_raw_monitor_to_present(
            ctx, raw_resource=before, idem_resource_name=name
        )
        result["comment"] = (f"Would delete aws.costexplorer.anomaly_monitor {name}",)
        return result
    else:
        result[
            "old_state"
        ] = hub.tool.aws.costexplorer.conversion_utils.convert_raw_monitor_to_present(
            ctx, raw_resource=before, idem_resource_name=name
        )

        ret = await hub.exec.boto3.client.ce.delete_anomaly_monitor(
            ctx, MonitorArn=resource_id
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = (f"Deleted aws.costexplorer.anomaly_monitor '{name}'",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns a list of aws cost anomaly monitors. If a monitor arn or name is specified, the list contains only
    the description of that monitor.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.costexplorer.anomaly_monitor
    """
    result = {}
    ret = await hub.exec.boto3.client.ce.get_anomaly_monitors(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe cost anomaly monitors {ret['comment']}")
        return {}

    for monitor in ret["ret"]["AnomalyMonitors"]:
        resource_id = monitor.get("MonitorArn")
        resource_translated = (
            hub.tool.aws.costexplorer.conversion_utils.convert_raw_monitor_to_present(
                ctx, raw_resource=monitor, idem_resource_name=resource_id
            )
        )
        result[resource_id] = {
            "aws.costexplorer.anomaly_monitor.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }
    return result



import copy
import re
from typing import Any
from typing import Dict
from typing import List


def decompose_resource_id(hub, resource_id: str) -> Dict[str, str]:
    """
    Route53 records ID is composed of zone identifier, record name, and record type, separated by underscores (_)

    Args:
        hub: The redistributed pop central hub.
        resource_id: String formatted as <hosted_zone_id>_<record_name>_<record_type>

    Returns:
        Dict containing the different parts of the ID.
    """
    m = re.match(r"(?P<zone_id>[A-Z0-9]+)_(?P<name>\S+)_(?P<type>[A-Z]+)", resource_id)
    if not m:
        raise Exception(f"{resource_id} does not match resource_id pattern")
    (hosted_zone_id, record_name, record_type) = (
        m.group("zone_id"),
        m.group("name"),
        m.group("type"),
    )
    return {
        "hosted_zone_id": f"/hostedzone/{hosted_zone_id}",
        "name": record_name,
        "record_type": record_type,
    }


def compose_resource_id(
    hub,
    hosted_zone_id: str,
    name: str,
    record_type: str,
) -> str:
    """
    Compose ID for a resource record in a hosted zone.

    Args:
        hub: The redistributed pop central hub.
        hosted_zone_id: The ID if the hosted zone containing this record.
        name: The name of the record.
        record_type: The record type.

    Returns:
        ID of this record.
    """
    prefix = "/hostedzone/"
    zone_id = hosted_zone_id
    l_name = name
    if hosted_zone_id.startswith(prefix):
        zone_id = hosted_zone_id[len(prefix) :]
    return f"""{zone_id}_{l_name}_{record_type}"""


def patch_state(
    hub,
    before: Dict[str, Any],
    hosted_zone_id: str,
    ttl: int,
    resource_records: List,
    alias_target: Dict,
) -> Dict[str, Any]:
    """
    Patches the current state of the resource to match the desired state.
    hosted_zone_id, name and record_type are not considered because are used as ID for this resource.

    Args:
        hub: The redistributed pop central hub.
        before (Dict): raw state of the resource record.
        hosted_zone_id (str): ID of the hosted zone of this resource record.
        ttl (int): TTL property of resource record. See boto3 or aws documentation for details.
        resource_records (List): Resource record set. See boto3 or aws documentation for details.
        alias_target (Dict): Alias target for alias resource record set. See boto3 or aws documentation for details.

    Returns:
        Patched resource record. Hosted zone id, name and record type are part of the ID of this resource, so they are
        immutable.
    """

    ret = hub.tool.aws.route53.conversion_utils.convert_raw_resource_record_to_present(
        hosted_zone_id, before
    )
    if ttl:
        ret["ttl"] = ttl
    if resource_records:
        ret["resource_records"] = resource_records
    if alias_target:
        ret["alias_target"] = alias_target

    return ret


def create_change_batch_for_update(hub, action, resource_record) -> Any:
    """
    Create an object that will be passed to ChangeBatch argument of
    `change_resource_record_sets https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/route53.html#Route53.Client.change_resource_record_sets`__

    Args:
        hub: The redistributed pop central hub.
        action(Text): 'CREATE'|'DELETE'|'UPSERT' literal.
        resource_record(Any): Idem resource record instance.

    Returns:
        Object to pass to ChangeBatch populated with the supported resource record properties
    """

    change_batch = {
        "Comment": f"{action} by idem-aws",
        "Changes": [
            {
                "Action": action,
                "ResourceRecordSet": {
                    "Name": resource_record["name"],
                    "Type": resource_record["record_type"],
                },
            },
        ],
    }
    if "ttl" in resource_record:
        change_batch["Changes"][0]["ResourceRecordSet"]["TTL"] = resource_record["ttl"]
    if "resource_records" in resource_record:
        change_batch["Changes"][0]["ResourceRecordSet"]["ResourceRecords"] = [
            {"Value": v} for v in resource_record["resource_records"]
        ]
    if "alias_target" in resource_record:
        change_batch["Changes"][0]["ResourceRecordSet"]["AliasTarget"] = {
            "HostedZoneId": resource_record["alias_target"]["hosted_zone_id"],
            "DNSName": resource_record["alias_target"]["dns_name"],
            "EvaluateTargetHealth": resource_record["alias_target"][
                "evaluate_target_health"
            ],
        }
    return change_batch


async def find_resource_record(
    hub, ctx, hosted_zone_id: str, name: str, record_type: str
) -> Dict[str, Any]:
    """
    Find a resource record in a hosted zone matching it by name and record type

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run.
        hosted_zone_id: ID of the hosted zone holding the resource record.
        name: name of the resource record.
        record_type: type of the resource record.

    Returns:
        Resource record matching the hosted_zone_id, name and record_type or None if resource record is not found.
    """

    record_sets = await hub.exec.boto3.client.route53.list_resource_record_sets(
        ctx,
        HostedZoneId=hosted_zone_id,
        StartRecordName=name,
        StartRecordType=record_type,
        MaxItems="1",
    )

    # present works on single resource record while `list_resource_record_sets` returns a list
    # expectation is if the resource record exists to be the first in the list
    if (
        record_sets["result"]
        and record_sets["ret"]["ResourceRecordSets"]
        and record_sets["ret"]["ResourceRecordSets"][0]["Name"] == name
        and record_sets["ret"]["ResourceRecordSets"][0]["Type"] == record_type
    ):
        return dict(
            result=record_sets["result"],
            comment=record_sets["comment"],
            aws_state=record_sets["ret"]["ResourceRecordSets"][0],
        )
    else:
        return dict(result=record_sets["result"], comment=record_sets["comment"])


def same_states(hub, state1: Dict[str, Any], state2: Dict[str, Any]) -> bool:
    """
    Compare Idem states for aws.route53.resource_record
    Args:
        state1(Dict[str,Any]): First state to compare
        state2(Dict[str,Any]): Second state to compare

    Returns:
        True if both states are equal, False otherwise
    """
    if state1 is None or state2 is None:
        return state1 == state2

    state1_copy = copy.deepcopy(state1)
    state2_copy = copy.deepcopy(state2)

    if "resource_records" in state1_copy and "resource_records" in state2_copy:
        state1_copy["resource_records"].sort()
        state2_copy["resource_records"].sort()
    # hosted_zone_id can come in 2 formats - /hostedzone/<id> and <id>. Convert to common format before comparison
    if not state1_copy["hosted_zone_id"].startswith("/hostedzone/"):
        state1_copy["hosted_zone_id"] = f"/hostedzone/{state1_copy['hosted_zone_id']}"
    if not state2_copy["hosted_zone_id"].startswith("/hostedzone/"):
        state2_copy["hosted_zone_id"] = f"/hostedzone/{state2_copy['hosted_zone_id']}"
    return state1_copy == state2_copy



import copy
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    rest_api_id: str,
    parent_id: str,
    path_part: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    """
    Creates a new resource or modifies an existing one.

    Args:
        name(string): An idem name of the resource.
        rest_api_id(string): AWS rest_api id of the associated RestApi.
        parent_id(string): The parent resource's id.
        path_part(string): The last path segment for this resource.
        resource_id(string, optional): AWS Resource id. Defaults to None.

    Request Syntax:
        [idem_test_aws_apigateway_resource]:
          aws.apigateway.resource.present:
            - name: 'string'
            - rest_api_id: 'string'
            - parent_id: 'string'
            - path_part: 'string'

    Returns:
        Dict[str, Any]

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False

    if resource_id:
        before = await hub.exec.aws.apigateway.resource.get(
            ctx, name=name, resource_id=resource_id, rest_api_id=rest_api_id
        )
        if not before["result"]:
            result["comment"] = before["comment"]
            result["result"] = False
            return result

    if before and before["ret"]:
        result[
            "old_state"
        ] = hub.tool.aws.apigateway.resource.convert_raw_resource_to_present(
            before["ret"], idem_resource_name=name, rest_api_id=rest_api_id
        )

        result["new_state"] = copy.deepcopy(result["old_state"])
        update_parameters = dict(
            {
                "parent_id": parent_id,
                "path_part": path_part,
            }
        )

        update_resource_ret = await hub.exec.aws.apigateway.resource.update_resource(
            ctx,
            old_state=result["old_state"],
            update_parameters=update_parameters,
        )

        if not update_resource_ret["result"]:
            result["result"] = False
            result["comment"] = update_resource_ret["comment"]
            return result
        result["comment"] = result["comment"] + update_resource_ret["comment"]
        resource_updated = bool(update_resource_ret["ret"])

        if resource_updated:
            if ctx.get("test", False):
                result["comment"] = result[
                    "comment"
                ] + hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.apigateway.resource", name=name
                )
                result["new_state"].update(update_resource_ret["ret"])
                return result

    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "rest_api_id": rest_api_id,
                    "parent_id": parent_id,
                    "path_part": path_part,
                },
            )

            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.apigateway.resource", name=name
            )
            return result

        ret = await hub.exec.boto3.client.apigateway.create_resource(
            ctx, restApiId=rest_api_id, parentId=parent_id, pathPart=path_part
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        resource_id = ret["ret"]["id"]
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.apigateway.resource", name=name
        )

    if (not before) or resource_updated:
        after = await hub.exec.boto3.client.apigateway.get_resource(
            ctx, resourceId=resource_id, restApiId=rest_api_id
        )
        if not (after["result"] and after["ret"]):
            result["result"] = False
            result["comment"] = result["comment"] + after["comment"]
            return result
        resource_translated = (
            hub.tool.aws.apigateway.resource.convert_raw_resource_to_present(
                after["ret"], idem_resource_name=name, rest_api_id=rest_api_id
            )
        )
        result["new_state"] = resource_translated

    return result


async def absent(
    hub, ctx, name: str, rest_api_id: str, resource_id: str = None
) -> Dict[str, Any]:
    r"""
    Deletes the specified resource.

    Args:
        name(string): The Idem name of the resource.
        rest_api_id(string): AWS rest_api id of the associated RestApi.
        resource_id(string, optional): AWS Resource id. Defaults to None.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.apigateway.resource.absent:
                - name: value
                - resource_id: value
                - rest_api_id: value


    """
    result = dict(
        comment=(),
        old_state=None,
        new_state=None,
        name=name,
        result=True,
    )

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigateway.resource", name=name
        )
        return result

    before_ret = await hub.exec.boto3.client.apigateway.get_resource(
        ctx, resourceId=resource_id, restApiId=rest_api_id
    )

    if not before_ret["result"]:
        if "NotFoundException" not in str(before_ret["comment"][0]):
            result["result"] = False
            result["comment"] = before_ret["comment"]
        else:
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.apigateway.resource", name=name
            )
        return result

    result[
        "old_state"
    ] = hub.tool.aws.apigateway.resource.convert_raw_resource_to_present(
        raw_resource=before_ret["ret"],
        rest_api_id=rest_api_id,
    )
    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.apigateway.resource", name=name
        )
        return result

    ret = await hub.exec.boto3.client.apigateway.delete_resource(
        ctx, restApiId=rest_api_id, resourceId=resource_id
    )

    if not ret["result"]:
        result["result"] = False
        result["comment"] = ret["comment"]
        return result

    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
        resource_type="aws.apigateway.resource", name=name
    )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""

    Describe the API Gateway Resources associated with a specific Rest API.

    Returns a list of apigateway.resource descriptions

    Returns:
        Dict[str, Any]


    Examples:

        .. code-block:: bash

            $ idem describe aws.apigateway.resource

    """

    result = {}

    get_rest_apis_ret = await hub.exec.boto3.client.apigateway.get_rest_apis(ctx)
    if not get_rest_apis_ret["result"]:
        hub.log.debug(f"Could not get Rest Apis {get_rest_apis_ret['comment']}")
        return result

    for rest_api in get_rest_apis_ret["ret"]["items"]:
        rest_api_name = rest_api.get("name")
        rest_api_id = rest_api.get("id")

        get_resources_ret = await hub.exec.boto3.client.apigateway.get_resources(
            ctx, restApiId=rest_api_id
        )
        if not get_resources_ret["result"]:
            hub.log.debug(
                f"Could not get Resources for Rest Api '{rest_api_name}': "
                f"{get_resources_ret['comment']}. Describe will skip this Rest Api and continue."
            )
            continue

        for resource in get_resources_ret["ret"]["items"]:
            # the resource response will always have a dictionary of size 2 which contains
            # just the parentId and an empty name, and we should filter those out
            if len(resource) != 2:
                resource_translated = (
                    hub.tool.aws.apigateway.resource.convert_raw_resource_to_present(
                        raw_resource=resource,
                        rest_api_id=rest_api_id,
                    )
                )

                result[resource_translated["resource_id"]] = {
                    "aws.apigateway.resource.present": [
                        {parameter_key: parameter_value}
                        for parameter_key, parameter_value in resource_translated.items()
                    ]
                }

    return result



from typing import Any
from typing import Dict


async def get_policy_if_attached_to_user(
    hub, ctx, user_name: str, policy_arn: str
) -> Dict[str, Any]:
    """
    Check if a managed policy is attached to a user

    Args:
        hub: The redistributed pop central hub.
        ctx: A dict with the keys/values for the execution of the Idem run located in
        `hub.idem.RUNS[ctx['run_name']]`.
        user_name: The name (friendly name, not ARN) of the IAM user to attach the policy to.
         This parameter allows (through its regex pattern ) a string of characters consisting of upper and lowercase alphanumeric characters with no spaces.
         You can also include any of the following characters: _+=,.@-
        policy_arn(Text): The Amazon Resource Name (ARN) of the IAM policy you want to attach.

    Returns:
        {"result": True|False, "comment": "A message", "ret": None}
    """
    result = dict(comment="", result=False, ret=None)
    user_policies_list = await hub.exec.boto3.client.iam.list_attached_user_policies(
        ctx, UserName=user_name
    )
    if user_policies_list["result"]:
        attached_user_policies_list = user_policies_list["ret"].get("AttachedPolicies")
        if attached_user_policies_list:
            policy_arn_list = [
                policy.get("PolicyArn") for policy in attached_user_policies_list
            ]
            if policy_arn in policy_arn_list:
                result["result"] = True
                result["ret"] = {"PolicyArn": policy_arn}
    else:
        result["comment"] = user_policies_list["comment"]
    return result



import botocore.client
import botocore.docs.docstring


def parse(hub, client: "botocore.client.BaseClient", service: str, operation: str):
    function = getattr(client, operation)
    doc: botocore.docs.docstring.ClientMethodDocstring = function.__doc__
    docstring = hub.tool.format.html.parse(doc._gen_kwargs["method_description"])
    parameters = {}
    try:
        # TODO what does botocore expect?
        params = doc._gen_kwargs["operation_model"].input_shape.members
        required_params = doc._gen_kwargs[
            "operation_model"
        ].input_shape.required_members
        for p, data in params.items():
            # Unwrap TagSpecification and get tags
            if "TagSpecifications" in p:
                p = "tags"
                data = data.member.members["Tags"]
            parameters[p] = hub.pop_create.aws.param.parse(
                param=data, required=p in required_params
            )
    except AttributeError as e:
        hub.log.error(f"Error reading parameters for {service}.{operation}: {e}")
        parameters = {}
    try:
        return_type = hub.pop_create.aws.plugin.type(
            doc._gen_kwargs["operation_model"].output_shape.type_name
        )
    except AttributeError:
        return_type = None

    return_fields = {}
    if return_type == "structure" or return_type == "Dict":
        try:
            return_fields = next(
                iter(doc._gen_kwargs["operation_model"].output_shape.members.items())
            )[1].members
        except AttributeError:
            return_fields = {}

    ret = {
        "doc": "\n".join(hub.tool.format.wrap.wrap(docstring, width=112)),
        "params": parameters,
        "return_type": return_type,
        "hardcoded": {
            "service": service,
            "operation": operation,
            "return_fields": return_fields,
        },
    }

    return ret



import re

import boto3.docs.docstring
import boto3.exceptions
import boto3.session


__func_alias__ = {"type_": "type"}


def parse(hub, session: "boto3.session.Session", service: str):
    plugins = {}

    # Create the boto client that will be parsed for capabilities
    client = session.client(
        service_name=service,
        region_name=hub.OPT.pop_create.region,
        api_version=hub.OPT.pop_create.api_version,
    )
    operations = {}
    for op in client.meta.method_to_api_mapping:
        try:
            verb, resource = op.split("_", maxsplit=1)
            if re.match(rf"\w+[^aoius]s$", resource):
                resource = hub.tool.format.inflect.singular(resource)
            # Special case for resource names that end with apis
            if resource.endswith("apis"):
                resource = resource[:-1]
            if resource not in operations:
                operations[resource] = {}
            operations[resource][verb] = op
        except ValueError:
            ...

    # Get resources if we can for this object
    resources = {}
    try:
        service_resource = session.resource(
            service_name=service,
            region_name=hub.OPT.pop_create.region,
            api_version=hub.OPT.pop_create.api_version,
        )
        for sub_resource in service_resource.meta.resource_model.subresources:
            resources[hub.tool.format.case.snake(sub_resource.name)] = (
                sub_resource.name,
                [],
            )
            resources[hub.tool.format.case.snake(sub_resource.name)][1].extend(
                [
                    f"hub.tool.boto3.resource.exec(resource, {action.name}, *args, **kwargs)"
                    for action in sub_resource.resource.model.actions
                ]
            )
    except boto3.exceptions.ResourceNotExistsError:
        ...

    clean_service = hub.tool.format.keyword.unclash(service).replace("-", "_")
    if clean_service != service.replace("-", "_"):
        plugin_docstring = hub.tool.format.html.parse(
            client._service_model.documentation
        )
        plugins[f"{clean_service}.init"] = {
            "imports": [],
            "functions": {},
            "doc": "\n".join(hub.tool.format.wrap.wrap(plugin_docstring, width=120)),
            "sub_alias": [service.replace("-", "_"), clean_service],
        }

    # Create a state for everything that has a create/delete/describe function
    for resource, functions in operations.items():
        other_calls = [
            f"hub.exec.boto3.client.{clean_service}.{op}" for op in functions.values()
        ]
        get_resource_call = ""

        if not resource:
            continue
        # Get resource and describe func
        if resource not in resources:
            # Prefer using the client first
            for func_name in hub.pop_create.aws.template.DESCRIBE_FUNCTIONS:
                if func_name in functions:
                    describe_function_call = f"await hub.exec.boto3.client.{clean_service}.{functions[func_name]}(resource_id)"
                    break
            else:
                hub.log.info(
                    f"Cannot determine how to describe {clean_service}.{resource}: {list(functions.keys())}"
                )
                describe_function_call = (
                    "await hub.tool.boto3.resource.describe(resource)"
                )
        else:
            # If the client wasn't complete, try to use a resource
            r = resources[resource]
            get_resource_call = f'resource = hub.tool.boto3.resource.create(ctx, "{clean_service}", "{r[0]}", resource_id)'
            other_calls.append(get_resource_call)
            describe_function_call = "await hub.tool.boto3.resource.describe(resource)"
            other_calls.extend(r[1])

        state_functions = {}

        # Get create function
        for func_name in hub.pop_create.aws.template.CREATE_FUNCTIONS:
            if func_name in functions:
                state_functions["present"] = functions[func_name]
                break
        else:
            hub.log.info(
                f"Cannot determine how to create {clean_service}.{resource}: {list(functions.keys())}"
            )

        # Get delete function
        for func_name in hub.pop_create.aws.template.DELETE_FUNCTIONS:
            if func_name in functions:
                state_functions["absent"] = functions[func_name]
                break
        else:
            hub.log.info(
                f"Cannot determine how to delete {clean_service}.{resource}: {list(functions.keys())}"
            )

        # Get list function
        for func_name in functions:
            if client.can_paginate(functions[func_name]):
                state_functions["describe"] = functions[func_name]
                break
        else:
            hub.log.info(
                f"Cannot determine how to describe {clean_service}.{resource}: {list(functions.keys())}"
            )

        # Skip resource if present function cannot be determined
        if not state_functions.get("present", None):
            continue

        clean_resource = hub.tool.format.keyword.unclash(resource)
        plugin_key = f"{clean_service}.{clean_resource}".replace("-", "_")
        plugins[plugin_key] = {
            "imports": [
                "import copy",
                "from dataclasses import field",
                "from dataclasses import make_dataclass",
                "from typing import *",
                "__contracts__ = ['resource']",
            ],
            "functions": {},
            "doc": str(client.__doc__),
        }
        if clean_resource != resource:
            plugins[plugin_key]["virtualname"] = resource

        plugins[plugin_key]["doc"] = "\n".join(other_calls)
        shared_function_data = {
            "delete_function": f"hub.exec.boto3.client.{clean_service}.{state_functions.get('absent', 'delete_function')}",
            "create_function": f"hub.exec.boto3.client.{clean_service}.{state_functions.get('present', 'create_function')}",
            "list_function": f"hub.exec.boto3.client.{clean_service}.{state_functions.get('describe', 'list_function')}",
            "waiter_call": "",
            "resource_function_call": get_resource_call,
            "describe_function_call": describe_function_call,
            "list_item": "TODO",
            "resource_id": "TODOs",
            "service_name": clean_service,
            "resource": resource,
            "has_client_token": False,
            "is_idempotent": False,
            "tag_method": "TODO, unify the way resources are tagged",
        }
        for state_function in ["present", "absent", "describe"]:
            if not state_functions.get(state_function, None):
                func_definition = {
                    "doc": "Missing function",
                    "params": {},
                    "return_type": None,
                    "hardcoded": {},
                }
            else:
                func_definition = hub.pop_create.aws.function.parse(
                    client, service, state_functions[state_function]
                )
            func_definition["hardcoded"].update(shared_function_data)
            func_definition["hardcoded"].update(
                {
                    "has_client_token": func_definition["params"].pop(
                        "ClientToken", None
                    ),
                    "is_idempotent": func_definition["params"].pop(
                        "ClientToken", "idempotent" in func_definition["doc"].lower()
                    ),
                }
            )

            # Normalize the name parameter
            if "Name" in func_definition["params"]:
                name = func_definition["params"].pop("Name")
            elif "name" in func_definition["params"]:
                name = func_definition["params"].pop("name")
            else:
                name = hub.pop_create.aws.template.NAME_PARAMETER.copy()
                if func_definition["hardcoded"]["is_idempotent"]:
                    name["doc"] = "The name of the state"

            resource_id_parameter = (
                hub.pop_create.aws.template.RESOURCE_ID_PARAMETER.copy()
            )
            if state_function == "absent":
                # resource_id parameter is required for absent() function
                resource_id_parameter["required"] = True
            # Create a new param list the puts "name" first in the ordered dict
            func_definition["params"] = dict(
                Name=name,
                resource_id=resource_id_parameter,
                **func_definition["params"],
            )
            plugins[plugin_key]["functions"][state_function] = func_definition

        # TODO this is where stuff needs to happen
        plugins[plugin_key]["functions"]["describe"]["hardcoded"]["present_params"] = [
            {k: "TODO"} for k in plugins[plugin_key]["functions"]["present"]["params"]
        ]

    return plugins



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.ec2.create_vpc_endpoint
hub.exec.boto3.client.ec2.delete_vpc_endpoints
hub.exec.boto3.client.ec2.describe_vpc_endpoints
hub.exec.boto3.client.ec2.modify_vpc_endpoint
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    vpc_id: str,
    service_name: str,
    resource_id: str = None,
    vpc_endpoint_type: str = None,
    policy_document: str = None,
    route_table_ids: List[str] = None,
    subnet_ids: List[str] = None,
    security_group_ids: List[str] = None,
    private_dns_enabled: bool = None,
    client_token: str = None,
    tags: List[Dict[str, Any]] or Dict[str, Any] = None,
    timeout: make_dataclass(
        "Timeout",
        [
            (
                "create",
                make_dataclass(
                    "CreateTimeout",
                    [
                        ("delay", int, field(default=0)),
                        ("max_attempts", int, field(default=0)),
                    ],
                ),
                field(default=None),
            ),
            (
                "update",
                make_dataclass(
                    "UpdateTimeout",
                    [
                        ("delay", int, field(default=0)),
                        ("max_attempts", int, field(default=0)),
                    ],
                ),
                field(default=None),
            ),
        ],
    ) = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a VPC endpoint for a specified service. An endpoint enables you to create a private connection between
    your VPC and the service. The service may be provided by Amazon Web Services, an Amazon Web Services Marketplace
    Partner, or another Amazon Web Services account. For more information, see VPC Endpoints in the Amazon Virtual
    Private Cloud User Guide. A gateway endpoint serves as a target for a route in your route table for traffic
    destined for the Amazon Web Service. You can specify an endpoint policy to attach to the endpoint, which will
    control access to the service from your VPC. You can also specify the VPC route tables that use the endpoint. An
    interface endpoint is a network interface in your subnet that serves as an endpoint for communicating with the
    specified service. You can specify the subnets in which to create an endpoint, and the security groups to
    associate with the endpoint network interface. A GatewayLoadBalancer endpoint is a network interface in your
    subnet that serves an endpoint for communicating with a Gateway Load Balancer that you've configured as a VPC
    endpoint service. Use DescribeVpcEndpointServices to get a list of supported services.

    Args:
        name(Text): An Idem name of the resource.
        vpc_id(Text): AWS VPC ID.
        resource_id(Text, optional): AWS VPC Endpoint id.
        service_name(Text): The service name. To get a list of available services, use the DescribeVpcEndpointServices
            request, or get the name from the service provider.
        vpc_endpoint_type(Text, optional): The type of endpoint. Default: Gateway. Defaults to Gateway.
        policy_document(Text, optional): (Interface and gateway endpoints) A policy to attach to the endpoint that controls access to the
            service. The policy must be in valid JSON format. If this parameter is not specified, we attach
            a default policy that allows full access to the service. Defaults to None.
        route_table_ids(List, optional): (Gateway endpoint) One or more route table IDs. Defaults to None.
        subnet_ids(List, optional): (Interface and Gateway Load Balancer endpoints) The ID of one or more subnets in which to create
            an endpoint network interface. For a Gateway Load Balancer endpoint, you can specify one subnet
            only. Defaults to None.
        security_group_ids(List, optional): (Interface endpoint) The ID of one or more security groups to associate with the endpoint
            network interface. Defaults to None.
        private_dns_enabled(bool, optional): (Interface endpoint) Indicates whether to associate a private hosted zone with the specified
            VPC. The private hosted zone contains a record set for the default public DNS name for the
            service for the Region (for example, kinesis.us-east-1.amazonaws.com), which resolves to the
            private IP addresses of the endpoint network interfaces in the VPC. This enables you to make
            requests to the default public DNS name for the service instead of the public DNS names that are
            automatically generated by the VPC endpoint service. To use a private hosted zone, you must set
            the following VPC attributes to true: enableDnsHostnames and enableDnsSupport. Use
            ModifyVpcAttribute to set the VPC attributes. Default: true. Defaults to None.
        client_token(Text, optional): Unique, case-sensitive identifier that you provide to ensure the idempotency of the request. Constraint: Maximum 64 ASCII characters. This field is autopopulated if not provided.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the VPC endpoint.
            Each tag consists of a key name and an associated value. Defaults to None.
            * (Key, optional): The key of the tag. Constraints: Tag keys are case-sensitive and accept a maximum of 127 Unicode
                characters. May not begin with aws:.
            * (Value, optional): The value of the tag. Constraints: Tag values are case-sensitive and accept a maximum of 256
                Unicode characters.
        timeout(Dict, optional): Timeout configuration for create/update/deletion of AWS IAM Policy.
            * create (Dict): Timeout configuration for creating AWS IAM Policy
                * delay (int, Optional): The amount of time in seconds to wait between attempts.
                * max_attempts (int, Optional): Customized timeout configuration containing delay and max attempts.
            * update(Dict, optional): Timeout configuration for updating AWS IAM Policy
                * delay (int, Optional): The amount of time in seconds to wait between attempts.
                * max_attempts: (int, Optional) Customized timeout configuration containing delay and max attempts.

    Request Syntax:
        [vpc_endpoint-id]:
            aws.ec2.vpc_endpoint.present:
            - vpc_id: 'string'
            - service_name: 'string'
            - resource_id: 'string'
            - vpc_endpoint_type: 'string'
            - policy_document: 'string'
            - route_table_ids:
              - 'string'
            - subnet_ids:
              - 'string'
            - security_group_ids:
              - 'string'
            - private_dns_enabled: 'Boolean'
            - client_token: 'string'
            - tag_specifications:
            - Key: 'string'
                Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            vpce-1231231234:
                aws.ec2.vpc_endpoint.present:
                - vpc_id: vpc-123123123
                - service_name: com.amazonaws.us-west-2.s3
                - resource_id: vpce-1231231234
                - vpc_endpoint_type: Interface
                - private_dns_enabled: False
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    if resource_id:
        before = await hub.exec.boto3.client.ec2.describe_vpc_endpoints(
            ctx, VpcEndpointIds=[resource_id]
        )
        if not before["result"]:
            result["result"] = False
            result["comment"] = result["comment"] + before["comment"]
            return result

    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if before and before["ret"].get("VpcEndpoints"):
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_vpc_endpoint_to_present(
            raw_resource=before["ret"]["VpcEndpoints"][0], idem_resource_name=name
        )
        old_tags = result["old_state"].get("tags", None)
        plan_state = copy.deepcopy(result["old_state"])
        if tags is not None and tags != old_tags:
            update_ret = await hub.exec.aws.ec2.tag.update_tags(
                ctx=ctx,
                resource_id=resource_id,
                old_tags=old_tags,
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_ret["comment"]
            if not update_ret["result"]:
                result["result"] = False
                return result
            if ctx.get("test", False) and update_ret["result"]:
                plan_state["tags"] = update_ret["ret"]

        hub.log.debug(f"Updating existing aws.ec2.vpc_endpoint '{name}'")
        ret = await hub.exec.aws.ec2.vpc_endpoint.update_vpc_endpoint(
            ctx,
            result["old_state"],
            resource_id=resource_id,
            policy_document=policy_document,
            route_table_ids=route_table_ids,
            subnet_ids=subnet_ids,
            security_group_ids=security_group_ids,
            private_dns_enabled=private_dns_enabled,
            timeout=timeout,
        )
        result["result"] = ret["result"]
        result["comment"] = ret["comment"]

        if not result["result"]:
            return result
        if ctx.get("test", False):
            result["new_state"] = ret["ret"]
            if plan_state.get("tags") is not None:
                result["new_state"]["tags"] = plan_state["tags"]
            return result
    else:
        if ctx.get("test", False):
            plan_state = {
                "name": name,
                "vpc_id": vpc_id,
                "service_name": service_name,
                "vpc_endpoint_type": vpc_endpoint_type,
                "policy_document": policy_document,
                "route_table_ids": route_table_ids,
                "subnet_ids": subnet_ids,
                "security_group_ids": security_group_ids,
                "private_dns_enabled": private_dns_enabled,
                "client_token": client_token,
                "tags": tags,
            }
            result["new_state"] = {k: v for k, v in plan_state.items() if v is not None}
            result["comment"] = (
                f"Would create aws.ec2.vpc_endpoint '{name}' and attach to vpc '{vpc_id}'",
            )
            return result

        hub.log.debug(f"Creating new aws.ec2.vpc_endpoint '{name}'")
        ret = await hub.exec.boto3.client.ec2.create_vpc_endpoint(
            ctx,
            ClientToken=client_token,
            VpcEndpointType=vpc_endpoint_type,
            VpcId=vpc_id,
            ServiceName=service_name,
            PolicyDocument=policy_document,
            RouteTableIds=route_table_ids,
            SubnetIds=subnet_ids,
            SecurityGroupIds=security_group_ids,
            PrivateDnsEnabled=private_dns_enabled,
            TagSpecifications=[
                {
                    "ResourceType": "vpc-endpoint",
                    "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
                }
            ]
            if tags
            else None,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        resource_id = ret["ret"]["VpcEndpoint"]["VpcEndpointId"]
        hub.log.debug(
            f"Created new aws.ec2.vpc_endpoint '{name}' with resource_id: '{resource_id}'"
        )
        result["comment"] = (
            f"Created aws.ec2.vpc_endpoint: '{name}' attached to vpc: '{vpc_id}'",
        )
    try:
        after = await hub.exec.boto3.client.ec2.describe_vpc_endpoints(
            ctx, VpcEndpointIds=[resource_id]
        )
        result[
            "new_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_vpc_endpoint_to_present(
            raw_resource=after["ret"]["VpcEndpoints"][0], idem_resource_name=name
        )
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub, ctx, name: str, resource_id: str = None, timeout: Dict = None
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified VPC endpoint. You can delete any of the following types of VPC endpoints.
    Gateway endpoint,   Gateway Load Balancer endpoint,   Interface endpoint   The following rules apply when you
    delete a VPC endpoint:   When you delete a gateway endpoint, we delete the endpoint routes in the route tables
    that are associated with the endpoint.   When you delete a Gateway Load Balancer endpoint, we delete the
    endpoint network interfaces.  You can only delete Gateway Load Balancer endpoints when the routes that are
    associated with the endpoint are deleted.   When you delete an interface endpoint, we delete the endpoint
    network interfaces.

    Args:
        name(Text): An idem name of the resource
        resource_id(Text, optional): An AWS vpc_endpoint resource_id to identify the resource.
        timeout(Dict, optional): Timeout configuration for deleting the endpoint.
            * delete (Dict) -- Timeout configuration for deleting the endpoint
                * delay -- The amount of time in seconds to wait between attempts.
                * max_attempts -- Customized timeout configuration containing delay and max attempts.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            vpce-123123123:
              aws.ec2.vpc_endpoint.absent:
                - name: value
                - resource_id: vpce-123123123
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_already_absent_msg = hub.tool.aws.comment_utils.already_absent_comment(
        resource_type="aws.ec2.vpc_endpoint", name=name
    )
    if not resource_id:
        result["comment"] = resource_already_absent_msg
        return result
    before = await hub.exec.boto3.client.ec2.describe_vpc_endpoints(
        ctx, VpcEndpointIds=[resource_id]
    )
    hub.log.debug(f"Attempting to delete aws.ec2.vpc_endpoint '{name}'")

    if not (before["ret"] and before["result"]):
        result["comment"] = resource_already_absent_msg
        return result
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_vpc_endpoint_to_present(
            raw_resource=before["ret"]["VpcEndpoints"][0], idem_resource_name=name
        )
        result["comment"] = (f"Would delete aws.ec2.vpc_endpoint '{name}'",)
        return result
    else:
        if before["ret"]["VpcEndpoints"][0]["State"] == "deleted":
            result["comment"] = resource_already_absent_msg
            return result
        ret = await hub.exec.boto3.client.ec2.delete_vpc_endpoints(
            ctx, VpcEndpointIds=[resource_id]
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + (ret["comment"],)
            return result
        result["comment"] = result["comment"] + (f"Deleted '{name}'",)

        delete_waiter_acceptors = [
            {
                "matcher": "pathAll",
                "expected": "available",
                "state": "retry",
                "argument": "VpcEndpoints[].State",
            },
            {
                "matcher": "pathAll",
                "expected": "pending",
                "state": "retry",
                "argument": "VpcEndpoints[].State",
            },
            {
                "matcher": "pathAll",
                "expected": "deleted",
                "state": "success",
                "argument": "VpcEndpoints[].State",
            },
            {
                "matcher": "error",
                "expected": "InvalidVpcEndpointId.NotFound",
                "state": "success",
                "argument": "Error.Code",
            },
        ]
        waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
            default_delay=15,
            default_max_attempts=40,
            timeout_config=timeout.get("delete") if timeout else None,
        )
        endpoint_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
            name="VpcEndpointDelete",
            operation="DescribeVpcEndpoints",
            argument=["Error.Code", "VpcEndpoints[].State"],
            acceptors=delete_waiter_acceptors,
            client=hub.tool.boto3.client.get_client(ctx, "ec2"),
        )
        try:
            await hub.tool.boto3.client.wait(
                ctx,
                "ec2",
                "VpcEndpointDelete",
                endpoint_waiter,
                VpcEndpointIds=[resource_id],
                WaiterConfig=waiter_config,
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False
    if before["result"] and before["ret"].get("VpcEndpoints"):
        result[
            "old_state"
        ] = hub.tool.aws.ec2.conversion_utils.convert_raw_vpc_endpoint_to_present(
            raw_resource=before["ret"]["VpcEndpoints"][0], idem_resource_name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Describes one or more of your VPC endpoints.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.ec2.vpc_endpoint
    """

    result = {}
    ret = await hub.exec.boto3.client.ec2.describe_vpc_endpoints(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe aws.ec2.vpc_endpoint {ret['comment']}")
        return result

    for vpc_endpoint in ret["ret"]["VpcEndpoints"]:
        resource_id = vpc_endpoint.get("VpcEndpointId")
        resource_translated = (
            hub.tool.aws.ec2.conversion_utils.convert_raw_vpc_endpoint_to_present(
                vpc_endpoint, idem_resource_name=resource_id
            )
        )
        result[resource_id] = {
            "aws.ec2.vpc_endpoint.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result



def is_pending(hub, ret: dict, state: str = None) -> bool:
    """
    This method enables state specific implementation of is_pending logic,
    based on resource specific attribute(s).
    Usage 'idem state <sls-file> --reconciler=basic --pending=aws'

    :param hub: The hub
    :param ret: The returned dictionary of the last run
    :param state: The name of the state
    :return: True | False
    """

    if (
        state is not None
        and hub.states[state] is not None
        and callable(getattr(hub.states[state], "is_pending", None))
    ):
        return hub.states[state].is_pending(ret=ret)
    else:
        # Default is_pending
        return not ret["result"] is True or bool(ret["changes"])



import copy
from typing import Any
from typing import Dict


def generate_test_state(
    hub,
    enforced_state: Dict[str, Any],
    desired_state: Dict[str, Any],
):
    """
    Giving the previous enforced state and desired state inputs, generate a test state,
     which can be used by idem state --test as the running result.
    If resource_id is not specified, auto-populate it.
    None values in desired state should be overridden by enforced state.

    Args:
        hub: The redistributed pop central hub.
        enforced_state: Previous enforced state.
        desired_state: A dictionary of desired state values. If any property's value is None,
         this property will be ignored. This is to match the behavior when a present() input is a None, Idem does not
         do an update.

    Returns:
        A result state.
    """
    desired_state_clean = {k: v for k, v in desired_state.items() if v is not None}
    if enforced_state:
        plan_state = copy.deepcopy(enforced_state)
        for parameter in desired_state_clean.keys():
            if desired_state_clean.get(parameter) is not None and plan_state.get(
                parameter
            ) != desired_state_clean.get(parameter):
                plan_state[parameter] = desired_state_clean.get(parameter)
    else:
        plan_state = copy.deepcopy(desired_state_clean)

    # For test we should add resource_id
    if plan_state.get("resource_id", None) is None:
        plan_state["resource_id"] = "resource_id_known_after_present"
    return plan_state



import copy
import re
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

delete_waiter_acceptors = [
    {
        "matcher": "error",
        "expected": "NoSuchEntity",
        "state": "retry",
        "argument": "Error.Code",
    },
    {
        "matcher": "path",
        "expected": "SUCCEEDED",
        "state": "success",
        "argument": "Status",
    },
    {
        "matcher": "path",
        "expected": "FAILED",
        "state": "success",
        "argument": "Status",
    },
    {
        "matcher": "path",
        "expected": "IN_PROGRESS",
        "state": "retry",
        "argument": "Status",
    },
    {
        "matcher": "path",
        "expected": "NOT_STARTED",
        "state": "retry",
        "argument": "Status",
    },
]


async def present(
    hub,
    ctx,
    name: str,
    service_name: str,
    custom_suffix: str = None,
    resource_id: str = None,
    description: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates an IAM role that is linked to a specific Amazon Web Services service. The service controls the attached policies
    and when the role can be deleted. This helps ensure that the service is not broken by an unexpectedly changed or deleted
    role, which could put your Amazon Web Services resources into an unknown state.

    Note:
        1. Updates to role name, custom suffix and description are not allowed by AWS. Only tags can be updated.
        2. All service linked roles does not support custom suffix

    Args:
        name(Text): The name of the idem resource.
        service_name(Text): The service principal for the Amazon Web Services service to which this role is attached.
                            for example: elasticbeanstalk.amazonaws.com
        custom_suffix(Text, Optional): A string that you provide, which is combined with the service-provided prefix to
                                        form the complete service linked role name
        resource_id(Text, Optional): AWS IAM Role Name.
        description(Text, Optional): A description of the service linked role. Defaults to None.
        tags(Dict or List, optional): Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the new service linked role.
            Each tag consists of a key name and an associated value. Defaults to None.
            * (Key): The key name that can be used to look up or retrieve the associated value. For example,
                Department or Cost Center are common choices.
            * (Value): The value associated with this tag. For example, tags with a key name of Department could have
                values such as Human Resources, Accounting, and Support. Tags with a key name of Cost Center
                might have values that consist of the number associated with the different cost centers in your
                company. Typically, many resources have tags with the same key name but with different values.
                Amazon Web Services always interprets the tag Value as a single string. If you need to store an
                array, you can store comma-separated values in the string. However, you must interpret the value
                in your code.
    Request Syntax:
        [iam-service-linked-role-name]:
          aws.iam.service_linked_role.present:
          - name: 'string'
          - service_name: 'string'
          - custom_suffix: 'string'
          - resource_id: 'string'
          - description: 'string'
          - tags:
            - Key: 'string'
              Value: 'string'

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls
            AWSServiceRoleForAutoScaling:
              aws.iam.service_linked_role.present:
                - name: AWSServiceRoleForAutoScaling
                - resource_id: AWSServiceRoleForAutoScaling
                - service_name: autoscaling.amazonaws.com
                - custom_suffix: test_suffix
                - description: This is custom description
                - tags:
                    - Key: firstkey
                      Value: firstvalue
                    - Key: 2ndkey
                      Value: 2ndvalue
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    resource = hub.tool.boto3.resource.create(
        ctx, "iam", "Role", resource_id if resource_id else name
    )
    before = await hub.tool.boto3.resource.describe(resource)
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
    if before:
        result[
            "old_state"
        ] = hub.tool.aws.iam.service_linked_role.convert_raw_service_linked_role_to_present(
            before
        )

        resource_updated = tags != result["old_state"].get("tags")
        result["new_state"] = copy.deepcopy(result["old_state"])
        if not resource_updated:
            result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
                resource_type="aws.iam.service_linked_role", name=name
            )
            return result
        else:
            update_ret = await hub.exec.aws.iam.role.update_role_tags(
                ctx,
                role_name=resource_id,
                old_tags=result["old_state"].get("tags"),
                new_tags=tags,
            )
            result["result"] = update_ret["result"]
            result["new_state"]["tags"] = tags
            result["comment"] = update_ret["comment"]
            return result

    if ctx.get("test", False):
        result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
            enforced_state={},
            desired_state={
                "name": name,
                "service_name": service_name,
                "custom_suffix": custom_suffix,
                "description": description,
                "resource_id": resource_id if resource_id else name,
                "tags": tags,
            },
        )
        operation = "update" if before and resource_updated else "create"
        result["comment"] = (f"Would {operation} aws.iam.service_linked_role '{name}'",)
        return result
    else:
        create_role_ret = await hub.exec.boto3.client.iam.create_service_linked_role(
            ctx=ctx,
            AWSServiceName=service_name,
            Description=description,
            CustomSuffix=custom_suffix,
        )
        if not create_role_ret["result"]:
            result["comment"] = create_role_ret["comment"]
            result["result"] = False
            return result
        role_name = create_role_ret["ret"]["Role"].get("RoleName")
        if tags:
            add_tag_ret = await hub.exec.boto3.client.iam.tag_role(
                ctx,
                RoleName=role_name,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
            )
            if not add_tag_ret["result"]:
                result["comment"] = add_tag_ret["comment"]
                result["result"] = False
                return result

        created_resource = hub.tool.boto3.resource.create(ctx, "iam", "Role", role_name)
        after_ret = await hub.tool.boto3.resource.describe(created_resource)
        result[
            "new_state"
        ] = hub.tool.aws.iam.service_linked_role.convert_raw_service_linked_role_to_present(
            after_ret
        )
        if not before:
            result["comment"] = hub.tool.aws.comment_utils.create_comment(
                resource_type="aws.iam.service_linked_role", name=name
            )
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Submits a service-linked role deletion request and returns a DeletionTaskId , which can be used to check the status of the deletion.

    Args:
        name(Text): AWS IAM service linked Role Name.
        resource_id(Text, Optional): AWS IAM service linked Role Name. If not specified, Idem will use "name"
         parameter to identify the IAM service linked role on AWS.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls
            resource_is_absent:
              aws.iam.service_linked_role.absent:
                - resource_id: AWSServiceRoleForAutoScaling
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_id = resource_id if resource_id else name
    resource = hub.tool.boto3.resource.create(ctx, "iam", "Role", resource_id)
    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.iam.service_linked_role", name=name
        )
    elif ctx.get("test", False):
        result[
            "old_state"
        ] = hub.tool.aws.iam.service_linked_role.convert_raw_service_linked_role_to_present(
            before
        )
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.iam.service_linked_role", name=name
        )
        return result
    else:
        try:
            result[
                "old_state"
            ] = hub.tool.aws.iam.service_linked_role.convert_raw_service_linked_role_to_present(
                before
            )

            ret = await hub.exec.boto3.client.iam.delete_service_linked_role(
                ctx=ctx, RoleName=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            deletionTaskId = ret["ret"].get("DeletionTaskId")
            # Custom waiter for delete
            # TODO - Use Reconciliation Loop
            waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
                default_delay=15,
                default_max_attempts=40,
                timeout_config=timeout.get("delete") if timeout else None,
            )
            cluster_waiter = hub.tool.boto3.custom_waiter.waiter_wrapper(
                name="ServiceLinkedRoleDelete",
                operation="GetServiceLinkedRoleDeletionStatus",
                argument=["Status", "Error.Code"],
                acceptors=delete_waiter_acceptors,
                client=hub.tool.boto3.client.get_client(ctx, "iam"),
            )
            await hub.tool.boto3.client.wait(
                ctx,
                "iam",
                "ServiceLinkedRoleDelete",
                cluster_waiter,
                DeletionTaskId=deletionTaskId,
                WaiterConfig=waiter_config,
            )
            deletion_status_ret = (
                await hub.exec.boto3.client.iam.get_service_linked_role_deletion_status(
                    ctx=ctx, DeletionTaskId=deletionTaskId
                )
            )
            if deletion_status_ret["result"]:
                deletion_status = deletion_status_ret["ret"]["Status"]
                if deletion_status == "FAILED":
                    result["result"] = False
                    result["comment"] = result["comment"] + (
                        deletion_status_ret["ret"]["Reason"].get("Reason"),
                    )
                    return result
                elif deletion_status == "SUCCEEDED":
                    result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                        resource_type="aws.iam.service_linked_role", name=name
                    )
            else:
                result["result"] = False
                result["comment"] = result["comment"] + deletion_status_ret["comment"]
                return result
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Lists the IAM service linked roles. If there are none, the operation returns an empty list.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.iam.service_linked_role
    """
    result = {}
    ret = await hub.exec.boto3.client.iam.list_roles(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe service_linked_role {ret['comment']}")
        return {}

    for role in ret["ret"]["Roles"]:
        # By default list_roles provides all role
        # Service linked roles have path something like 'Path': '/aws-service-role/autoscaling.amazonaws.com/'
        # and other all roles have path as "/"
        if not re.search("/aws-service-role/*", str(role)):
            continue
        resource = hub.tool.boto3.resource.create(
            ctx, "iam", "Role", role.get("RoleName")
        )
        resource = await hub.tool.boto3.resource.describe(resource)
        translated_resource = hub.tool.aws.iam.service_linked_role.convert_raw_service_linked_role_to_present(
            resource
        )
        resource_key = f"iam-service_linked_role-{translated_resource['resource_id']}"
        result[resource_key] = {
            "aws.iam.service_linked_role.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }
    return result



from collections import OrderedDict
from typing import Dict
from typing import List


def convert_raw_policy_to_present(hub, policy: Dict, tags: List = None):
    """
    Convert organization policy response to a common format

    Args:
        policy: organization policy response from aws
        tags: list of tags attached to policy

    Returns:
        A dictionary of organization policy
    """
    translated_resource = {}
    describe_parameters = OrderedDict(
        {
            "Name": "name",
            "Description": "description",
            "Type": "policy_type",
            "Content": "content",
            "Id": "resource_id",
            "Arn": "arn",
            "AwsManaged": "aws_managed",
        }
    )

    for camel_case_key, snake_case_key in describe_parameters.items():
        if policy["PolicySummary"].get(camel_case_key) is not None:
            translated_resource[snake_case_key] = policy["PolicySummary"].get(
                camel_case_key
            )
        elif policy.get(camel_case_key) is not None:
            translated_resource[snake_case_key] = policy.get(camel_case_key)
    if tags:
        translated_resource["tags"] = tags
    return translated_resource


def convert_raw_organization_to_present(hub, organization: Dict, idem_name: str = None):
    """
    Convert the aws organization response to a common format

    Args:
        organization: describe organization response
        idem_name: Name of the resource

    Returns:
        A dictionary of organization
    """
    resource_id = organization.get("Id")
    translated_resource = {
        "name": idem_name if idem_name else resource_id,
        "resource_id": resource_id,
    }
    describe_parameters = OrderedDict(
        {
            "Arn": "arn",
            "MasterAccountArn": "master_account_arn",
            "MasterAccountId": "master_account_id",
            "MasterAccountEmail": "master_account_email",
            "FeatureSet": "feature_set",
            "AvailablePolicyTypes": "available_policy_types",
        }
    )

    if not "FeatureSet" in organization:
        organization["FeatureSet"] = "ALL"

    for camel_case_key, snake_case_key in describe_parameters.items():
        if organization.get(camel_case_key) is not None:
            translated_resource[snake_case_key] = organization.get(camel_case_key)
    return translated_resource


def convert_raw_ou_to_present(hub, ou: Dict, parent_id: str = None):
    """
    Convert organization policy response to a common format

    Args:
        ou: organization unit response from aws
        parent_id: id of the root

    Returns:
        A dictionary of organization unit
    """
    translated_resource = {"resource_id": ou.get("Id")}
    if ou.get("Name"):
        translated_resource["name"] = ou.get("Name")
    if parent_id:
        translated_resource["parent_id"] = parent_id

    if ou.get("Tags"):
        translated_resource["tags"] = ou.get("Tags")
    return translated_resource


def convert_raw_account_to_present(
    hub, parent_id: str = None, account: Dict = None, tags: Dict = None
):
    """
    Convert account response to a common format

    Args:
        hub: required for functions in hub
        parent_id: id of the account parent
        account: Account response from AWS
        tags: Tags of account

    Returns:
        A dictionary of account
    """
    describe_parameters = OrderedDict(
        {
            "Email": "email",
            "Arn": "arn",
            "Status": "status",
        }
    )
    resource_translated = {
        "name": account.get("Name"),
        "resource_id": account.get("Id"),
    }

    if parent_id:
        resource_translated["parent_id"] = parent_id
    for parameter_raw, parameter_present in describe_parameters.items():
        if parameter_raw in account:
            resource_translated[parameter_present] = account.get(parameter_raw)
    if tags:
        resource_translated["tags"] = tags
    return resource_translated



"""

hub.exec.boto3.client.events.delete_rule
hub.exec.boto3.client.events.describe_rule
hub.exec.boto3.client.events.disable_rule
hub.exec.boto3.client.events.enable_rule
hub.exec.boto3.client.events.list_rules
hub.exec.boto3.client.events.put_rule
"""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    targets: List = None,
    schedule_expression: str = None,
    event_pattern: str = None,
    rule_status: str = None,
    description: str = None,
    role_arn: str = None,
    event_bus_name: str = None,
    tags: List[Dict[str, Any]] or Dict[str, Any] = None,
) -> Dict[str, Any]:
    r"""

    Enables the specified rule. If the rule does not exist, the operation fails. When you enable a rule, incoming
    events might not immediately start matching to a newly enabled rule. Allow a short period of time for changes to
    take effect.

    Args:
        hub:
        ctx:
        name(Text): A name, ID to identify the resource.
        resource_id(Text, Optional): The name of the AWS CloudWatch Events Rule.
        targets(List): The targets to update or add to the rule.
        schedule_expression (Text, Optional): Scheduling expression.
            For example, "cron(0 20 * * ? *)" or "rate(5 minutes)".
        event_pattern (Text, Optional): Rules use event patterns to select events and route them to targets. A pattern
            either matches an event or it doesn't. Event patterns are represented as JSON objects with a structure that
            is similar to that of events.
        rule_status (Text, Optional): Indicates whether the rule is enabled or disabled.
        description (Text, Optional): A description of the rule.
        role_arn (Text, Optional): The Amazon Resource Name (ARN) of the IAM role associated with the rule.
            If you're setting an event bus in another account as the target and that account granted permission to your
            account through an organization instead of directly by the account ID, you must specify a RoleArn with proper
            permissions in the Target structure, instead of here in this parameter.
        event_bus_name (Text, Optional) : The name or ARN of the event bus to associate with this rule.
            If you omit this, the default event bus is used.
        tags(List or Dict, optional): list of tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
            {tag-key: tag-value} The tags to assign to the rule. Defaults to None.

    Request Syntax:
        [cloudwatch-events-rule-id]:
          aws.events.rule.present:
            - name: event_rule
            - arn: arn:aws:events:us-west-1:000000000000:rule/qqqqqqq
            - rule_status: ENABLED | DISABLED
            - schedule_expression: rate(5 minutes)
            - event_bus_name: default
            - tags:
              - Key: 'string'
                Value: 'string'


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            event_rule_id:
                aws.events.rule.present:
                  - name: event_rule
                  - arn: arn:aws:events:us-west-1:000000000000:rule/test
                  - rule_status: ENABLED | DISABLED
                  - schedule_expression: rate(5 minutes)
                  - event_bus_name: default
                  - tags:
                    - Key: idem_test_event_rule1
                      Value: test value

    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    are_tags_updated = False
    is_role_updated = False
    existing_tags = {}
    plan_state = {}
    rule_arn = None
    before = await hub.exec.boto3.client.events.describe_rule(ctx, Name=resource_id)
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)
    if before["result"] and before["ret"]:
        result["comment"] = (f"'{name}' already exists",)
        rule_arn = before["ret"]["Arn"]
        tags_return = await hub.exec.boto3.client.events.list_tags_for_resource(
            ctx, ResourceARN=rule_arn
        )
        result["result"] = tags_return.get("result")
        if result["result"]:
            existing_tags = tags_return.get("ret").get("Tags")
        else:
            result["comment"] = result["comment"] + tags_return["comment"]
            return result

        resource_converted = await hub.tool.aws.events.conversion_utils.convert_raw_cloud_watch_rule_to_present(
            ctx,
            raw_resource=before["ret"],
            idem_resource_name=name,
            tags=existing_tags,
        )
        result["result"] = result["result"] and resource_converted["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + resource_converted["comment"]
        result["old_state"] = resource_converted["ret"]
        plan_state = copy.deepcopy(result["old_state"])

        update_return = await hub.exec.aws.events.rule.update_events_rule(
            ctx=ctx,
            old_targets=result["old_state"].get("targets"),
            new_targets=targets,
            schedule_expression=schedule_expression,
            event_pattern=event_pattern,
            state=rule_status,
            plan_state=plan_state,
            role_arn=role_arn,
            description=description,
            event_bus_name=event_bus_name,
            resource_id=resource_id,
        )
        result["result"] = result["result"] and update_return["result"]
        result["comment"] = result["comment"] + update_return["comment"]
        if result["result"]:
            is_role_updated = True

        if tags is not None and tags != result["old_state"].get("tag"):
            # Update tags
            update_tags_result = await hub.exec.aws.events.tag.update_tags(
                ctx=ctx,
                resource_id=rule_arn,
                old_tags=result["old_state"].get("tag"),
                new_tags=tags,
            )
            result["comment"] = result["comment"] + update_tags_result["comment"]
            result["result"] = result["result"] and update_tags_result["result"]

            if not result["result"]:
                return result
            are_tags_updated = result["result"]

            if ctx.get("test", False) and (update_tags_result["ret"] is not None):
                plan_state["tags"] = update_tags_result["ret"]
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "targets": targets,
                    "tags": tags,
                    "resource_id": name,
                    "schedule_expression": schedule_expression,
                    "event_pattern": event_pattern,
                    "description": description,
                    "event_bus_name": event_bus_name,
                    "role_arn": role_arn,
                    "rule_status": rule_status,
                },
            )
            result["comment"] = (f"Would create aws.events.rule '{name}'",)
            return result
        try:
            ret = await hub.exec.boto3.client.events.put_rule(
                ctx=ctx,
                Name=name,
                ScheduleExpression=schedule_expression,
                EventPattern=event_pattern,
                State=rule_status,
                Description=description,
                RoleArn=role_arn,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                if tags
                else None,
                EventBusName=event_bus_name,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                return result
            if targets is not None:
                target_ret = await hub.exec.boto3.client.events.put_targets(
                    ctx=ctx,
                    Rule=name,
                    Targets=targets,
                    EventBusName=event_bus_name,
                )
                result["result"] = result["result"] and target_ret["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + ret["comment"]
                    return result
            resource_id = name
            rule_arn = ret["ret"]["RuleArn"]
            existing_tags = (
                hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags) if tags else []
            )
            result["comment"] = (f"Created aws.events.rule '{name}'",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)
            result["result"] = False

    if ctx.get("test", False):
        result["new_state"] = plan_state
    elif not (before and before["result"]) or is_role_updated:
        after = await hub.exec.boto3.client.events.describe_rule(ctx, Name=resource_id)
        final_updated_tags = existing_tags
        if are_tags_updated:
            tags = await hub.exec.boto3.client.events.list_tags_for_resource(
                ctx, ResourceARN=rule_arn
            )
            if tags.get("result") and tags.get("ret"):
                final_updated_tags = tags.get("ret").get("Tags")
        resource_converted = await hub.tool.aws.events.conversion_utils.convert_raw_cloud_watch_rule_to_present(
            ctx,
            raw_resource=after["ret"],
            idem_resource_name=name,
            tags=final_updated_tags,
        )
        result["result"] = result["result"] and resource_converted["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + resource_converted["comment"]
        result["new_state"] = resource_converted["ret"]
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str,
    event_bus_name: str = None,
) -> Dict[str, Any]:
    r"""

    Deletes the specified rule. Before you can delete the rule, you must remove all targets, using RemoveTargets.
    When you delete a rule, incoming events might continue to match to the deleted rule. Allow a short period of
    time for changes to take effect. If you call delete rule multiple times for the same rule, all calls will
    succeed. When you call delete rule for a non-existent custom eventbus, ResourceNotFoundException is returned.
    Managed rules are rules created and managed by another Amazon Web Services service on your behalf. These rules
    are created by those other Amazon Web Services to support functionality in those services. You can
    delete these rules using the Force option, but you should do so only if you are sure the other service is not
    still using that rule.

    Args:
        ctx:
        hub:
        resource_id(Text): The name of the AWS CloudWatch Events Rule.
        name(Text): A name, ID to identify the resource.
        event_bus_name (Text, Optional) : The name or ARN of the event bus to associate with this rule.
            If you omit this, the default event bus is used.

    Request Syntax:
    [rule-id]:
      aws.events.rule.absent:
      - name: 'string'
      - resource_id: 'string'
      - event_bus_name: 'string'
      - qualifier: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.events.rule.absent:
                - name: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = await hub.exec.boto3.client.events.describe_rule(ctx, Name=resource_id)
    if not before["result"]:
        result["comment"] = (f"aws.events.rule '{name}' already absent",)
    else:
        tags = await hub.exec.boto3.client.events.list_tags_for_resource(
            ctx, ResourceARN=before["ret"].get("Arn")
        )
        result["result"] = tags.get("result")
        if result["result"]:
            existing_tags = tags.get("ret").get("Tags")
        else:
            result["comment"] = result["comment"] + tags["comment"]
            return result

        resource_converted = await hub.tool.aws.events.conversion_utils.convert_raw_cloud_watch_rule_to_present(
            ctx,
            raw_resource=before["ret"],
            idem_resource_name=before["ret"].get("Name"),
            tags=existing_tags,
        )
        result["result"] = result["result"] and resource_converted["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + resource_converted["comment"]
        result["old_state"] = resource_converted["ret"]
        if ctx.get("test", False):
            result["comment"] = result["comment"] + (
                f"Would delete aws.events.rule '{name}'",
            )
            return result
        else:
            try:
                targets = await hub.exec.boto3.client.events.list_targets_by_rule(
                    ctx, Rule=name
                )
                result["result"] = result["result"] and targets["result"]
                if result["result"] and targets["ret"]["Targets"]:
                    ids = []
                    for target in targets["ret"]["Targets"]:
                        ids.append(target.get("Id"))
                    remove_targets = await hub.exec.boto3.client.events.remove_targets(
                        ctx, Rule=name, Ids=ids
                    )
                    result["result"] = result["result"] and remove_targets["result"]
                    if not result["result"]:
                        result["comment"] = (
                            result["comment"] + remove_targets["comment"]
                        )
                        return result
                else:
                    result["comment"] = result["comment"] + targets["comment"]
                    return result
                ret = await hub.exec.boto3.client.events.delete_rule(
                    ctx, Name=name, EventBusName=event_bus_name
                )
                result["result"] = ret["result"] and result["result"]
                if not result["result"]:
                    result["comment"] = result["comment"] + ret["comment"]
                    return result
                result["comment"] = result["comment"] + (
                    f"Deleted aws.events.rule '{name}'",
                )
            except hub.tool.boto3.exception.ClientError as e:
                result["comment"] = result["comment"] + (
                    f"{e.__class__.__name__}: {e}",
                )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Lists your Amazon EventBridge rules. You can either list all the rules or you can provide a prefix to match to
    the rule names. ListRules does not list the targets of a rule. To see the targets associated with a rule, use
    ListTargetsByRule.


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.events.rule
    """

    result = {}
    ret = await hub.exec.boto3.client.events.list_rules(ctx)
    if not ret["result"]:
        hub.log.debug(f"Could not describe aws.events.rule {ret['comment']}")
        return result
    for rule in ret["ret"]["Rules"]:
        tags = await hub.exec.boto3.client.events.list_tags_for_resource(
            ctx, ResourceARN=rule.get("Arn")
        )
        resource_converted = await hub.tool.aws.events.conversion_utils.convert_raw_cloud_watch_rule_to_present(
            ctx,
            raw_resource=rule,
            idem_resource_name=rule.get("Name"),
            tags=tags.get("ret").get("Tags") if tags.get("result") else None,
        )
        if not resource_converted["result"]:
            hub.log.warning(
                f"Could not describe aws.events.rule '{rule.get('Name')}' "
                f"with error {resource_converted['comment']}"
            )
        else:
            result[rule.get("Arn")] = {
                "aws.events.rule.present": [resource_converted["ret"]]
            }
    return result



from typing import Dict

from dict_tools import data


async def update_cluster(
    hub,
    ctx,
    name: str,
    before: Dict,
    version: str,
    encryption_config: Dict,
    resources_vpc_config: Dict,
    logging: Dict,
    client_request_token: str,
    timeout: Dict,
):
    """
    Updates an Amazon EKS cluster

    Args:
       hub:
       ctx:
       name(Text): An Idem name of the resource.
       before(Dict): AWS cluster
       logging(Dict): Enable or disable exporting the Kubernetes control plane logs for your cluster to
                                CloudWatch Logs. By default, cluster control plane logs aren't exported to CloudWatch Logs. For more
                                information, see Amazon EKS Cluster control plane logs in the * Amazon EKS User Guide *
       encryption_config(Dict): The encryption configuration for the cluster.
       resources_vpc_config(Dict): The VPC configuration that's used by the cluster control plane. Amazon
                                   EKS VPC resources have specific requirements to work properly with Kubernetes. For more information,
                                   see Cluster VPC Considerations and Cluster Security Group Considerations in the Amazon EKS User Guide . You
                                   must specify at least two subnets. You can specify up to five security groups. However, we recommend that you
                                   use a dedicated security group for your cluster control plane.
       version(Text): The desired Kubernetes version for your cluster
       client_request_token(Text):Unique, case-sensitive identifier that you provide to ensure the idempotency of the request.
                                  This field is autopopulated if not provided.
       timeout(Dict, optional): Timeout configuration for creating or updating cluster.
            * create (Dict) -- Timeout configuration for creating cluster
                * delay(int, default=60) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=40) -- Customized timeout configuration containing delay and max attempts.
            * update (string) -- Timeout configuration for updating cluster
                * delay(int, default=60) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=40) -- Customized timeout configuration containing delay and max attempts.

    Returns:
       {"result": True|False, "comment": A message Tuple, "ret": Dict}
    """
    result = dict(comment=(), result=True, ret={})
    # change depends on resourcesVpcConfig,  encryptionConfig, logging and version
    if version and before.get("version") != version:
        if not ctx.get("test", False):
            ret = await hub.exec.boto3.client.eks.update_cluster_version(
                ctx,
                name=before.get("resource_id"),
                version=version,
                clientRequestToken=client_request_token,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            else:
                waiter_ret = await hub.exec.aws.eks.cluster.cluster_waiter(
                    ctx, name, before.get("resource_id"), timeout, "update"
                )
                if not waiter_ret["result"]:
                    result["result"] = False
                    result["comment"] = result["comment"] + waiter_ret["comment"]
                    return result
        result["ret"]["version"] = version
        result["comment"] = result["comment"] + (
            f"Update cluster version: {version} on cluster {before.get('resource_id')}",
        )

    # we can associate encryption config only if its already not present.
    # If the cluster already has encryption config we cannot associate.
    if (
        encryption_config
        and not before.get("encryption_config")
        and before.get("encryption_config") != encryption_config
    ):
        if not ctx.get("test", False):
            ret = await hub.exec.boto3.client.eks.associate_encryption_config(
                ctx,
                clusterName=before.get("resource_id"),
                encryptionConfig=encryption_config,
                clientRequestToken=client_request_token,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            else:
                waiter_ret = await hub.exec.aws.eks.cluster.cluster_waiter(
                    ctx, name, before.get("resource_id"), timeout, "update"
                )
                if not waiter_ret["result"]:
                    result["result"] = False
                    result["comment"] = result["comment"] + waiter_ret["comment"]
                    return result
        result["ret"]["encryption_config"] = encryption_config
        result["comment"] = result["comment"] + (
            f"Update cluster encryption config: {encryption_config} on cluster {before.get('resource_id')}",
        )
    if logging and data.recursive_diff(
        before.get("logging"), logging, ignore_order=True
    ):
        if not ctx.get("test", False):
            ret = await hub.exec.boto3.client.eks.update_cluster_config(
                ctx,
                name=before.get("resource_id"),
                clientRequestToken=client_request_token,
                logging=logging,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            else:
                waiter_ret = await hub.exec.aws.eks.cluster.cluster_waiter(
                    ctx, name, before.get("resource_id"), timeout, "update"
                )
                if not waiter_ret["result"]:
                    result["result"] = False
                    result["comment"] = result["comment"] + waiter_ret["comment"]
                    return result
        result["ret"]["logging"] = logging
        result["comment"] = result["comment"] + (
            f"Update cluster logging: {logging} on cluster {before.get('resource_id')}",
        )
    if resources_vpc_config:
        vpc_config_changes = hub.tool.aws.eks.eks_utils.get_resource_vpc_config_changes(
            before.get("resources_vpc_config"),
            resources_vpc_config,
        )
        if vpc_config_changes:
            if not ctx.get("test", False):
                ret = await hub.exec.boto3.client.eks.update_cluster_config(
                    ctx,
                    name=before.get("resource_id"),
                    clientRequestToken=client_request_token,
                    resourcesVpcConfig=vpc_config_changes,
                )
                result["result"] = ret["result"]
                if not result["result"]:
                    result["comment"] = ret["comment"]
                    return result
                else:
                    waiter_ret = await hub.exec.aws.eks.cluster.cluster_waiter(
                        ctx, name, before.get("resource_id"), timeout, "update"
                    )
                    if not waiter_ret["result"]:
                        result["result"] = False
                        result["comment"] = result["comment"] + waiter_ret["comment"]
                        return result
            result["ret"]["resources_vpc_config"] = resources_vpc_config
            result["comment"] = result["comment"] + (
                f"Update cluster resources vpc config: {vpc_config_changes} on cluster {before.get('resource_id')}",
            )

    return result


async def cluster_waiter(
    hub, ctx, name: str, resource_id: str, timeout: Dict, operation_type: str
):
    """

    Waiter to wait for the cluster to become active.

        Args:
           hub:
           ctx:
           name(Text): An Idem name of the resource.
           resource_id(Text): Name of the cluster to Identify the resource
           timeout(Dict, optional): Timeout configuration for creating or updating cluster.
            * create (Dict) -- Timeout configuration for creating cluster
                * delay(int, default=60) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=40) -- Customized timeout configuration containing delay and max attempts.
            * update (string) -- Timeout configuration for updating cluster
                * delay(int, default=60) -- The amount of time in seconds to wait between attempts.
                * max_attempts(int, default=40) -- Customized timeout configuration containing delay and max attempts.
           operation_type(Text): create or update operation

        Returns:
            Dict[str, Any]
    """

    result = dict(comment=(), result=True, ret={})
    waiter_config = hub.tool.aws.waiter_utils.create_waiter_config(
        default_delay=60,
        default_max_attempts=40,
        timeout_config=timeout.get(operation_type) if timeout else None,
    )
    hub.log.debug(f"Waiting on {operation_type} aws.eks.cluster '{name}'")
    try:
        await hub.tool.boto3.client.wait(
            ctx,
            "eks",
            "cluster_active",
            name=resource_id,
            WaiterConfig=waiter_config,
        )
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result



import copy
from collections import OrderedDict
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    description: str = None,
    version: str = None,
    binary_media_types: List[str] = None,
    minimum_compression_size: int = None,
    api_key_source: str = None,
    endpoint_configuration: make_dataclass(
        "EndpointConfiguration",
        [
            ("types", List[str], field(default=None)),
            ("vpcEndpointIds", List[str], field(default=None)),
        ],
    ) = None,
    policy: str = None,
    disable_execute_api_endpoint: bool = None,
    tags: Dict[str, Any] = None,
) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Creates a new rest_api resource or modifies an existing one.

    Args:
        name(Text): An idem name of the resource.
        resource_id(Text, optional): AWS rest_api id. Defaults to None.
        description (Text, optional): The description of the RestApi .
        version (Text, optional): A version identifier for the API.
        binary_media_types (List[str], optional):  The list of binary media types supported by the RestApi . By default, the RestApi supports only UTF-8-encoded text payloads.
        minimum_compression_size (integer, optional):  A nullable integer that is used to enable compression (with non-negative between 0 and 10485760 (10M) bytes, inclusive) or disable compression (with a null value) on an API. When compression is enabled, compression or decompression is not applied on the payload if the payload size is smaller than this value. Setting it to zero allows compression for any payload size.
        api_key_source (Text, optional):  The source of the API key for metering requests according to a usage plan.
        endpoint_configuration(Dict, optional): The endpoint configuration of this RestApi showing the endpoint types of the API. Defaults to None.
            * (types, optional): A list of endpoint types of an API (RestApi) or its custom domain name (DomainName).
                For an edge-optimized API and its custom domain name, the endpoint type is "EDGE". For a regional API
                and its custom domain name, the endpoint type is REGIONAL. For a private API, the endpoint type
                is PRIVATE.
            * (vpcEndpointIds, optional): A list of VpcEndpointIds of an API (RestApi) against which to create Route53 ALIASes. It is only
                supported for PRIVATE endpoint type.
        policy (Text, optional):  A stringified JSON policy document that applies to this RestApi regardless of the caller and Method configuration.
        tags(Dict, optional): The key-value map of strings. The valid character set is [a-zA-Z+-=._:/]. The tag key can be up
            to 128 characters and must not start with aws:. The tag value can be up to 256 characters. Defaults to None.
        disable_execute_api_endpoint (boolean, optional):  Specifies whether clients can invoke your API by using the default execute-api endpoint. By default, clients can invoke your API with the default https://{api_id}.execute-api.{region}.amazonaws.com endpoint. To require that clients use a custom domain name to invoke your API, disable the default endpoint.

    Request Syntax:
        [idem_test_aws_apigateway_rest_api]:
          aws.apigateway.rest_api.present:
            - name: 'string'
            - resource_id: 'string'
            - api_key_source: 'string'
            - binary_media_types:
              - 'string'
            - endpoint_configuration:
                types:
                - REGIONAL | EDGE | PRIVATE
                vpcEndpointIds:
                - 'string'
            - tags:
                'string': 'string'
            - disable_execute_api_endpoint: True | False
            - minimum_compression_size: int(0 to 10485760)
            - policy: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_present:
              aws.apigateway.rest_api.present:
                -name: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    before = None
    resource_updated = False
    if resource_id:
        before = await hub.exec.boto3.client.apigateway.get_rest_api(
            ctx, restApiId=resource_id
        )
        if not before["result"]:
            result["comment"] = before["comment"]
            result["result"] = False
            return result
        result[
            "old_state"
        ] = hub.tool.aws.apigateway.rest_api.convert_raw_rest_api_to_present(
            before["ret"], idem_resource_name=name
        )
    if before and before["ret"]:
        result["new_state"] = copy.deepcopy(result["old_state"])
        resource_arn = (
            f"arn:aws:apigateway:{ctx['acct']['region_name']}::/restapis/{resource_id}"
        )
        updateable_resource_parameters = OrderedDict(
            {
                "description": description,
                "binary_media_types": binary_media_types,
                "minimum_compression_size": minimum_compression_size,
                "api_key_source": api_key_source,
                "endpoint_configuration": endpoint_configuration,
                "disable_execute_api_endpoint": disable_execute_api_endpoint,
            }
        )

        update_rest_api_ret = await hub.exec.aws.apigateway.rest_api.update_rest_api(
            ctx,
            old_state=result["old_state"],
            updateable_resource_parameters=updateable_resource_parameters,
        )
        if not update_rest_api_ret["result"]:
            result["result"] = False
            result["comment"] = update_rest_api_ret["comment"]
            return result
        result["comment"] = result["comment"] + update_rest_api_ret["comment"]
        resource_updated = bool(update_rest_api_ret["ret"])

        if resource_updated and ctx.get("test", False):
            result["new_state"].update(update_rest_api_ret["ret"])

        if tags is not None and tags != result["old_state"].get("tags", {}):
            update_tags_ret = await hub.exec.aws.apigateway.tag.update_tags(
                ctx,
                resource_arn=resource_arn,
                old_tags=result["old_state"].get("tags", {}),
                new_tags=tags,
            )
            if not update_tags_ret["result"]:
                result["result"] = False
                result["comment"] = update_tags_ret["comment"]
                return result
            result["comment"] = result["comment"] + update_tags_ret["comment"]

            resource_updated = resource_updated or bool(update_tags_ret["ret"])
            if update_tags_ret["ret"] and ctx.get("test", False):
                result["new_state"]["tags"] = update_tags_ret["ret"]

        if resource_updated:
            if ctx.get("test", False):
                result["comment"] = result[
                    "comment"
                ] + hub.tool.aws.comment_utils.would_update_comment(
                    resource_type="aws.apigateway.rest_api", name=name
                )
                return result
            else:
                result["comment"] = result[
                    "comment"
                ] + hub.tool.aws.comment_utils.update_comment(
                    resource_type="aws.apigateway.rest_api", name=name
                )
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "resource_id": resource_id,
                    "description": description,
                    "version": version,
                    "binary_media_types": binary_media_types,
                    "minimum_compression_size": minimum_compression_size,
                    "api_key_source": api_key_source,
                    "endpoint_configuration": endpoint_configuration,
                    "policy": policy,
                    "disable_execute_api_endpoint": disable_execute_api_endpoint,
                    "tags": tags,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.apigateway.rest_api", name=name
            )
            return result
        ret = await hub.exec.boto3.client.apigateway.create_rest_api(
            ctx,
            name=name,
            description=description,
            version=version,
            binaryMediaTypes=binary_media_types,
            minimumCompressionSize=minimum_compression_size,
            apiKeySource=api_key_source,
            endpointConfiguration=endpoint_configuration,
            policy=policy,
            tags=tags,
            disableExecuteApiEndpoint=disable_execute_api_endpoint,
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        resource_id = ret["ret"]["id"]
        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.apigateway.rest_api", name=name
        )
    if (not before) or resource_updated:
        after = await hub.exec.boto3.client.apigateway.get_rest_api(
            ctx, restApiId=resource_id
        )
        if not (after["result"] and after["ret"]):
            result["result"] = False
            result["comment"] = result["comment"] + after["comment"]
            return result
        resource_translated = (
            hub.tool.aws.apigateway.rest_api.convert_raw_rest_api_to_present(
                after["ret"], idem_resource_name=name
            )
        )
        result["new_state"] = resource_translated
    else:
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""
    **Autogenerated function**

    Deletes the specified rest_api resource.

    Args:
        name(Text): An Idem name of the resource.
        resource_id(Text): AWS rest_api id. Idem automatically considers this resource as absent if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            resource_is_absent:
              aws.apigateway.rest_api.absent:
                - name: value
                - resource_id: value
    """

    already_absent_msg = hub.tool.aws.comment_utils.already_absent_comment(
        resource_type="aws.apigateway.rest_api", name=name
    )
    result = dict(
        comment=already_absent_msg,
        old_state=None,
        new_state=None,
        name=name,
        result=True,
    )

    if not resource_id:
        return result

    before = await hub.exec.boto3.client.apigateway.get_rest_api(
        ctx, restApiId=resource_id
    )

    if not (before["result"] and before["ret"]):
        if not ("NotFoundException" in str(before["comment"])):
            result["comment"] = before["comment"]
            result["result"] = False
        return result

    resource_translated = (
        hub.tool.aws.apigateway.rest_api.convert_raw_rest_api_to_present(
            before["ret"],
            idem_resource_name=name,
        )
    )
    result["old_state"] = resource_translated

    if ctx.get("test", False):
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.apigateway.rest_api", name=name
        )
    else:
        ret = await hub.exec.boto3.client.apigateway.delete_rest_api(
            ctx, restApiId=resource_id
        )
        result["result"] = ret["result"]
        if not result["result"]:
            result["comment"] = ret["comment"]
            return result
        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.apigateway.rest_api", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    Returns a list of apigateway.rest_api descriptions

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe aws.apigateway.rest_api
    """

    result = {}
    ret = await hub.exec.boto3.client.apigateway.get_rest_apis(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe rest_api {ret['comment']}")
        return {}

    for rest_api in ret["ret"]["items"]:
        name = rest_api.get("name")
        resource_translated = (
            hub.tool.aws.apigateway.rest_api.convert_raw_rest_api_to_present(
                rest_api,
                idem_resource_name=name,
            )
        )
        result[name] = {
            "aws.apigateway.rest_api.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource_translated.items()
            ]
        }

    return result
